#ifndef _RNACONS_H_
#define _RNACONS_H_

#include "Util.h"
#include "RNAStruct.h"
#include "MSA.h"
#include <map>

//const double Bstar = -0.20;
//const double var1 = 1.0;
//const double var2 = 1.0;
class stem; 
class stemlist; 
class PartStackStruct;
class RNA;

class SCORE {
public: 
	double ec_score; //energy score + covariance score
	double e_score;  //energy score; 
	SCORE(){ec_score = e_score = 0;}
	SCORE(const double & d1, const double & d2) {
		ec_score = d1; e_score = d2; 
	}
	double getec(){return ec_score;} 
	double gete(){return e_score;}
	double getc(){return ec_score-e_score;}
	void show(ostream & fout=cout) {fout<<"("<<ec_score<<", "<<e_score<<")"<<endl;}
};

SCORE operator + (const SCORE & sc1, const SCORE & sc2);
SCORE operator - (const SCORE & sc1, const SCORE & sc2);
//Assume < and > is determined by the first variable ec
bool operator < (const SCORE & sc1, const SCORE & sc2);
bool operator > (const SCORE & sc1, const SCORE & sc2);
bool operator == (const SCORE & sc1, const SCORE & sc2); 

bool eq(const SCORE & sc1, const SCORE & sc2);

class RNACons{
public:
	MSA msa;
	int N; //number of sequences
	int L; //length of sequences
	
	vector<vector<vector<int>>> seq_pi;
	vector<vector<vector<vector<int>>>> d;//d_{i,j}^{alpha,beta}
	vector<vector<double>> c; //covariance matrix
	vector<vector<double>> q;
	vector<vector<double>> b;
	vector<vector<bool>> ali_pi;
	vector<vector<double>> e; //energy contribution for pairing column i and j
	
	vector<double> F; //F(j) the minimum free energy for msa[0..j]
	vector<double> FC; //FC(j) the minimum free energy + covariance
	vector<vector<double>> E;//E(i,j) the minimum free energy for msa[i..j]
	vector<vector<double>> EC;//the minimum free energy + covariance

	RNACons(){}
	RNACons(const RNACons & in_rc);
	RNACons(const MSA & in_msa);

	double getec(const int & i, const int & j);
	SCORE fold();
	string backtrace();
	string backtraceF(const int & j, const string & knownstruct);
	string backtraceE(const int & i, const int & j, const string & knownstruct);
	
	SCORE calscore(seglist & theta) ;
	SCORE calscore(seglistT & theta); 
	SCORE calscore(bplist  & P    ) ;
	SCORE calscore(PartStruct & phi);
	SCORE calscore(PartStructT & phi);
	int subopt(ostream & fout = cout, const double & delta = 1, const bool & is_countec = true);

	int countbp();
	void show_ali_pi();
	double gethydrogenbonds(const int & x, const int & y, const int & p);
	
	//Stack based consensus folding and subopt structure generation
	SCORE calscore(PartStackStruct & phi);
	SCORE stackfold(stemlist & in_stl);
	string stacktraceback(stemlist & in_stl, const bool & is_countec);
	vector<RNA> stacksubopt(stemlist & in_stl, ostream & fout = cout, const double & delta = 1, const bool & is_coutec = true);

	//Saturated folding and suboptimal structure generation
	vector<vector<SCORE>> P0; 
	vector<vector<SCORE>> P1; 
	map<BP, vector<BP>> FFI;
	
	void sat_init();
	void sat_compute_FFI();
	
	vector<BP> sat_getFFI(const int & i, const int & j);
	SCORE sat_fold();
	string sat_backtrace();
	void sat_backtrace(string & str, const int & i, const int & j, const bool & isP0);
	vector<RNA> sat_subopt(ostream & fout, const double & ec_ub);

private:
	void compute_seq_pi();
	void compute_d();
	void compute_c();
	void compute_q();
	void compute_b();
	void compute_ali_pi();
	void compute_e();
	
	void show_seq_pi();
	void show_d();
	void show_c();
	void show_q();
	void show_b();

	void show_e();
	void show_ec();
	
	double getE(const int & i, const int & j);
	double getEC(const int & i, const int & j);
	double getF(const int & j);
	double getFC(const int & j);
	SCORE getFSCORE(const int & j);
	SCORE getESCORE(const int & i, const int & j);
	SCORE computeSCORE(const stem & st);
	void clearFE();
	string stacktracebackE(string & str, stemlist & stl, const int & i, const int & j, const bool & is_countec);
};
	
void show(vector<vector<SCORE>> vvsc, ostream & fout = cout);

class RNASingle {
public:
	string seq; 
	int L; 
	
	vector<int> F;
	vector<vector<int>> E;
	RNASingle(const string & in_seq) {
		seq = in_seq; 
		L = sz(seq);
	}
	
	int fold();
	string backtrace();
	string backtraceF(const int & j, const string & knowstruct);
	string backtraceE(const int & i, const int & j, const string & knownstruct);
	int calscore(PartStruct & phi);
	int subopt(ostream & fout = cout, const int & delta = 1);
	int countbp();
private:
};


#endif
