#include "RNACons_Turner.h"
#include "Config.h"
extern Config cf; 
extern TurnerE TE;

/*******************************************************************************
                        CLASS SCORE
 ******************************************************************************/
SCORE operator + (const SCORE & sc1, const SCORE & sc2) {
	return SCORE(sc1.ec+sc2.ec, sc1.e+sc2.e);
}

SCORE operator - (const SCORE & sc1, const SCORE & sc2) {
	return SCORE(sc1.ec-sc2.ec, sc1.e-sc2.e);
}

//Assume < and > is determined by the first variable ec
bool operator < (const SCORE & sc1, const SCORE & sc2) {
	return (sc1.ec < sc2.ec || (sc1.ec==sc2.ec && sc1.e < sc2.e));
}
bool operator > (const SCORE & sc1, const SCORE & sc2) {
	return (sc1.ec > sc2.ec || (sc1.ec==sc2.ec && sc1.e > sc2.e));
}

bool operator == (const SCORE & sc1, const SCORE & sc2) {
	return (sc1.ec==sc2.ec && sc1.e==sc2.e);
}

/******************************************************************************
                          CLASS  StackFoldAuxiliary
 ******************************************************************************/
StackFoldAuxiliary::StackFoldAuxiliary() 
{
	ctype = ctype_unknown;
	st = stem(-1,-1,-1,-1);
}

StackFoldAuxiliary::StackFoldAuxiliary(const CTYPE & in_ctype, const stem & in_t) 
{
	set(in_ctype, in_t);
}

StackFoldAuxiliary::StackFoldAuxiliary(const StackFoldAuxiliary & in_sfa) {
	set(in_sfa.ctype, in_sfa.st);
}

void StackFoldAuxiliary::set(const CTYPE & in_ctype, const stem & in_t) {
	ctype = in_ctype;
	st = in_t;
}

void StackFoldAuxiliary::show(ostream & fout) {
	showCTYPE(ctype, fout);
	fout<<"\t("<<st.outer.lp<<","<<st.outer.rp<<")~("<<st.inner.lp<<","<<st.inner.rp<<")"<<endl;
}

void showCTYPE(const CTYPE & ctype, ostream & fout) {
	switch(ctype) {
		case ctype_unknown: {fout<<"unknown";break;}
		case ctype_0: {fout<<"0";break;}
		case ctype_1: {fout<<"1";break;}
		case ctype_2: {fout<<"2";break;}
		case ctype_3: {fout<<"3";break;}
		case ctype_4: {fout<<"4";break;}
		case ctype_1a: {fout<<"1.a";break;}
		case ctype_1b: {fout<<"1.b";break;}
		case ctype_1c: {fout<<"1.c";break;}
		case ctype_2a: {fout<<"2.a";break;}
		case ctype_2b: {fout<<"2.b";break;}
		case ctype_2c: {fout<<"2.c";break;}
		case ctype_3a: {fout<<"3.a";break;}
		case ctype_3b: {fout<<"3.b";break;}
		case ctype_3c: {fout<<"3.c";break;}
		case ctype_3d: {fout<<"3.d";break;}
		case ctype_3e: {fout<<"3.e";break;}
		case ctype_3f: {fout<<"3.f";break;}
		case ctype_3g: {fout<<"3.g";break;}
		case ctype_3h: {fout<<"3.h";break;}
		case ctype_3i: {fout<<"3.i";break;}
		case ctype_3p: {fout<<"3.p";break;}
		case ctype_3q: {fout<<"3.q";break;}
		case ctype_3r: {fout<<"3.r";break;}				
		case ctype_3s: {fout<<"3.s";break;}
		case ctype_4a: {fout<<"4.a";break;}
		case ctype_4b: {fout<<"4.b";break;}
		case ctype_5a: {fout<<"5.a";break;}
		case ctype_5b: {fout<<"5.b";break;}
		default:{cerr<<"unknown CTYPE="<<ctype<<endl;exit(0);}
	}
}
/*******************************************************************************
                        CLASS RNACONST
 ******************************************************************************/
//constructor, copy from in_rc
RNAConsT::RNAConsT(const RNAConsT & in_rc) {
	msa = in_rc.msa;
	N = in_rc.N;
	L = in_rc.L;
	c = in_rc.c;
	ali_pi = in_rc.ali_pi;
}

//constructor, construct a new RNAConsT from in_msa
RNAConsT::RNAConsT(const MSA & in_msa) {
	msa = in_msa;
	N = in_msa.N;
	L = in_msa.L;
}

/* Initialize c and ali_pi
 */
void RNAConsT::init() {
	//initilize bases (-ACGU);
	base['_'] = 0; base['A'] = 1; base['C'] = 2; base['G'] = 3; base['U'] = 4;
	//initilize pairing pattern
	pair = {
	/*  -  A  C  G  U  */
		{0, 0, 0, 0, 0 }, //-
		{0, 0, 0, 0, 5 }, //A
		{0, 0, 0, 1, 0 }, //C
		{0, 0, 2, 0, 3 }, //G
		{0, 6, 0, 4, 0 }  //U
	};
	
	//initilize the corvariance matrix c and the consensus pairing matrix ali_pi
	FOR(i,0,L) {
		vector<int> vi;
		vector<bool> vb;
		FOR(j,0,L) {
			vi.PB(-OO);
			vb.PB(false);
		}
		c.PB(vi);
		ali_pi.PB(vb);
	}
	
	//initilize C, FM, M1, M2, M4, F2n2d, FnFd, F
	FOR(i, 0, L) {
		F.PB(+OO);
		Fn.PB(+OO);
		Fd.PB(+OO);
		F2n.PB(+OO);
		F2d.PB(+OO);
		FOR(j, i, L) {
			int indx = pos(i,j);
			C.PB(+OO);
			assert(sz(C)-1==indx);
			FM.PB(+OO);
			M1.PB(+OO);
			M2.PB(+OO);
			M4.PB(+OO);
		}
	}
	//initialize F_sfa, Fn_sfa, Fd_sfa, F2n_sfa, F2d_sfa
	//initialize C_sfa, FM_sfa, M1_sfa, M2_sfa, M4_sfa;
	StackFoldAuxiliary sfa;
	FOR(i, 0, L) {
		F_sfa.PB(sfa);
		Fn_sfa.PB(sfa);
		Fd_sfa.PB(sfa);
		F2n_sfa.PB(sfa);
		F2d_sfa.PB(sfa);
		FOR(j, i, L) {
			int indx = pos(i,j);
			C_sfa.PB(sfa);
			assert(indx == sz(C_sfa)-1);
			FM_sfa.PB(sfa);
			M1_sfa.PB(sfa);
			M2_sfa.PB(sfa);
			M4_sfa.PB(sfa);
		}
	}	
	
	//initilize F_S, Fn_S, Fd_S, F2n_S, F2d_S
	//initilize FM_LS, FM_RS, M1_LS, M1_RS, M2_LS, M2_RS, M4_LS, M4_RS
	stem s(-1, -1, -1, -1);
	FOR(i, 0, L) {
		F_S.PB(s); 
		Fn_S.PB(s);
		Fd_S.PB(s);
		F2n_S.PB(s);
		F2d_S.PB(s);
		FOR(j, i, L) {
			FM_LS.PB(s);
			FM_RS.PB(s);
			M1_LS.PB(s);
			M1_RS.PB(s);
			M2_LS.PB(s);
			M2_RS.PB(s);
			M4_LS.PB(s);
			M4_RS.PB(s);
			assert(sz(FM_LS)-1 == pos(i,j));
		}
	}
}

/* c[i][j] = covariance bonus for cell (i,j), 0 - conserved pairs, > 0 for good pairs*/
void RNAConsT::compute_covariance() {
	//Initialize the covariance matrix
	//CG = 1 GC = 2 GU =  3 UG = 4 AU = 5 UA = 6 gapgap = 7
	//hamming distance between base pairs
	
	int hd[7][7] = {
//    --  CG GC GU UG AU UA
		{0, 0, 0, 0, 0, 0, 0}, //--
		{0, 0, 2, 2, 1, 2, 2}, //CG
		{0, 2, 0, 1, 2, 2, 2}, //GC
		{0, 2, 1, 0, 2, 1, 2}, //GU
		{0, 1, 2, 2, 0, 2, 1}, //UG
		{0, 2, 2, 1, 2, 0, 2}, //AU
		{0, 2, 2, 2, 1, 2, 0}, //UA
	};
	
	FOR(i, 0, L) {
		FOR(j, i+cf.MINLOOPLEN+1, L) {
			int numBPOcc[8] = {0,0,0,0,0,0,0,0};
			FOR(k, 0, N) {
				int basepair; 
				if(msa[k][i] == '-' && msa[k][j] == '-') basepair = 7;
				else {
					basepair = pair[base[msa[k][i]]][base[msa[k][j]]];
				}
				numBPOcc[basepair]++; 
			}
			if(numBPOcc[0]*2 + numBPOcc[7] > N) {
				c[i][j] = -OO; continue; 
			}
			int score = 0;
			FOR(k, 1, 7) {
				FOR(l, k, 7) {
					score += numBPOcc[k] * numBPOcc[l]* hd[k][l]; 
				}
			}
			c[i][j] = cf.VAR1 * ((UNIT*score)/N-cf.VAR2*UNIT*(numBPOcc[0] + numBPOcc[7]*0.25));
			if(c[i][j] >= UNIT*cf.BSTAR) {
//				cout<<"c[i][j] = "<<c[i][j] <<" >= "<<UNIT*cf.BSTAR<<endl;
				ali_pi[i][j] = true;
			}
		}
	}
	
}

void RNAConsT::init_stl() {
	stl = stemlist(ali_pi, msa);
}

void RNAConsT::show_c() {
	FOR(i,0,L) {
		FOR(j,0,L) {
			if(c[i][j] == -OO) {cout<<"X\t";}
			else cout<<c[i][j]<<"\t";
		}
		cout<<endl;
	}
}

int RNAConsT::pos(const int & i, const int & j) {
	return (2*L-i+1)*i/2+j-i;
}
	
int RNAConsT::stackfold() {
	
	FOR(k, cf.MINLOOPLEN+1, L) {
		int i = 0; int j = i+k;
		if(j<0 || j>=L) break; 
		while(i>=0 && i<=L-1 && j>=0 && j<=L-1) {
			fillC (i, j);
			fillFM(i, j);
			fillM1(i, j);
			fillM2(i, j);
			fillM4(i, j);
			i++; j++;
		}
	}
	
	FOR(i, 0, L) {
		fillF2n2d(i);
		fillFnFd(i);
		fillF(i);
	}
	return F[L-1];
}


void RNAConsT::fillC(const int & i, const int & j) {
	stem s0 = stl.inwhichST(i, j);
	if(!s0.valid() || (s0.valid() && ((i-s0.outer.lp)>cf.MAXOVERLAP || s0.inner.lp-i+1<2))) return;
	
	stem s = stl.getsubstem_startfrom_endat(i, j);
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	assert(s.valid());
	assert(i==sx && j==sy);
	
	//case C.(1)
	int e = computeH(msa, s);
	int ij = pos(i,j);
	C_sfa[ij].ctype = ctype_1;
	
	vector<stem> vst = stl.getsubstems_enclosedby_or_inward_extend(s);
	FOR(k, 0, sz(vst)) {
		stem t = vst[k];
		int tx = t.outer.lp; int ty = t.outer.rp;
		int tp = t.inner.lp; int tq = t.inner.rp;
		assert(!s.conflictwith(t));
		
		//case C.(2)
		int e1 = computeS(msa, s) + computephi(msa, s, t) + C[pos(tx,ty)];
		if(e1 < e) {
			e = e1;
			C_sfa[ij].set(ctype_2, t);
		}

		int delta = computeS(msa, s) + N*(TE.Mc + 2*TE.Mphi) +
		            TP(msa, sp, sq) + C[pos(tx,ty)] + TP(msa, tx, ty);

		if(ty==sq-1) {
		} else if(ty==sq-2) {
			delta += min(d(msa, tx, ty, ty+1), d(msa, sp, sq, sq-1));
		} else if(ty<=sq-3) {
			delta += d(msa, tx, ty, ty+1) + d(msa, sp, sq, sq-1);
		}
		int i1=sp+1, i2=sp+2, i3=sp+3;
		int j1=tx-1, j2=tx-2, j3=tx-3;
		
		//case C.(3.a)
		int eij = getE(etype_M4, i1, j1);
		int e3a = delta + eij;
		if(e3a < e && eij != OO) {
			e = e3a;
			C_sfa[ij].set(ctype_3a, t);
		}
		
		//case C.(3.b)
		stem w = getStem(M4_RS, i1, j2); //M4_RS[i1][j2];
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_M4, i1, j2);
			int e3b = delta + eij
			          + min(d(msa, wx, wy, wy+1), d(msa, tx, ty, tx-1));
			if(e3b < e && eij != OO) {
				e = e3b;
				C_sfa[ij].set(ctype_3b, t);
			}
  		}
  		
  		//case C.(3.c)
  		w = getStem(M4_RS, i1, j3); //M4_RS[i1][j3];
  		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_M4, i1, j3);
			int e3c = delta + eij
			          + d(msa, wx, wy, wy+1) + d(msa, tx, ty, tx-1);
			if(e3c < e && eij != OO) {
				e = e3c;
				C_sfa[ij].set(ctype_3c, t);
			}
		}
		
		//case C.(3.d)
		stem v = getStem(M4_LS, i2, j1); //M4_LS[i2][j1];
		if(v.valid()) {
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M4, i2, j1);
			int e3d = delta+ eij
			          + min(d(msa, sp, sq, sp+1), d(msa, vx, vy, vx-1));
			if(e3d < e && eij != OO) {
				e = e3d;
				C_sfa[ij].set(ctype_3d, t);
			}
		}

		//case C.(3.e)
		v = getStem(M4_LS, i2, j2); //M4_LS[i2][j2];
		w = getStem(M4_RS, i2, j2); //M4_RS[i2][j2];
		if(v.valid() && w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M4, i2, j2);
			int e3e = delta + eij
			          + min(d(msa, sp, sq, sp+1), d(msa, vx, vy, vx-1))
			          + min(d(msa, wx, wy, wy+1), d(msa, tx, ty, tx-1));

			if(e3e < e && eij != OO) {
				e = e3e;
				C_sfa[ij].set(ctype_3e, t);
			}
		}
		
		//case C.(3.f)
		v = getStem(M4_LS, i2, j3); //M4_LS[i2][j3];
		w = getStem(M4_RS, i2, j3); //M4_RS[i2][j3];
		if(v.valid() && w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M4, i2, j3);
			int e3f = delta + eij
			          + min(d(msa, sp, sq, sp+1), d(msa, vx, vy, vx-1)) 
			          + d(msa, wx, wy, wy+1) + d(msa, tx, ty, tx-1);

			if(e3f < e && eij != OO) {
				e = e3f;
				C_sfa[ij].set(ctype_3f, t);
			}
		}
		
		//case C.(3.g)
		v = getStem(M4_LS, i3, j1); //M4_LS[i3][j1];
		if(v.valid()) {
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M4, i3, j1);
			int e3g = delta + eij
			          + d(msa, sp, sq, sp+1) + d(msa, vx, vy, vx-1);

			if(e3g < e && eij != OO) {
				e = e3g;
				C_sfa[ij].set(ctype_3g, t);
			}
		}
		
		//case C.(3.h)
		w = getStem(M4_RS, i3, j2); //M4_RS[i3][j2];
		v = getStem(M4_LS, i3, j2); //M4_LS[i3][j2];
		if(v.valid() && w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M4, i3, j2);
			int e3h = delta + eij
			          + d(msa, sp, sq, sp+1) + d(msa, vx, vy, vx-1)
			          + min(d(msa, wx, wy, wy+1), d(msa, tx, ty, tx-1));

			if(e3h < e && eij != OO) {
				e = e3h;
				C_sfa[ij].set(ctype_3h, t);
			}
		}
//Start: find out a bug here
		w = getStem(M2_RS, i3, j1); //M2_RS[i3][j1]
		v = getStem(M2_LS, i3, j1); //M2_LS[i3][j1]
		if(v.valid() && w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M2, i3, j1);
			int e3p = delta + eij + 
			          + d(msa, sp, sq, sp+1) + d(msa, vx, vy, vx-1);

			if(e3p < e && eij != OO) {
				e = e3p;
				C_sfa[ij].set(ctype_3p, t);
			}
		}
		
		w = getStem(M2_RS, i3, j2); //M2_RS[i3][j2]
		v = getStem(M2_LS, i3, j2); //M2_LS[i3][j2]
		if(v.valid() && w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M2, i3, j2);
			int e3q = delta + eij + 
			          + d(msa, sp, sq, sp+1) + d(msa, vx, vy, vx-1)
			          + min(d(msa, wx, wy, wy+1), d(msa, tx, ty, tx-1));

			if(e3q < e && eij != OO) {
				e = e3q;
				C_sfa[ij].set(ctype_3q, t);
			}
		}
		
		w = getStem(M1_RS, i1, j3); //M1_RS[i1][j3]
		v = getStem(M1_LS, i1, j3); //M1_LS[i1][j3]
		if(v.valid() && w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M1, i1, j3);
			int e3r = delta + eij + 
			          + d(msa, wx, wy, wy+1) + d(msa, tx, ty, tx-1);

			if(e3r < e && eij != OO) {
				e = e3r;
				C_sfa[ij].set(ctype_3r, t);
			}
		}
		
		w = getStem(M1_RS, i2, j3); //M1_RS[i1][j3]
		v = getStem(M1_LS, i2, j3); //M1_LS[i1][j3]
		if(v.valid() && w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_M1, i2, j3);
			int e3s = delta + eij + 
						 + min(d(msa, sp, sq, sp+1), d(msa, vx, vy, vx-1)) 
			          + d(msa, wx, wy, wy+1) + d(msa, tx, ty, tx-1);

			if(e3s < e && eij != OO) {
				e = e3s;
				C_sfa[ij].set(ctype_3s, t);
			}
		}
		
//End: find a bug here
		//case C.(3.i)
		w = getStem(FM_RS, i3, j3); //FM_RS[i3][j3];
		v = getStem(FM_LS, i3, j3); //FM_LS[i3][j3];
		if(v.valid() && w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int vx = v.outer.lp; int vy = v.outer.rp;
			eij = getE(etype_FM, i3, j3);
			int e3i = delta + eij
			          + d(msa, sp, sq, sp+1) + d(msa, vx, vy, vx-1)
			          + d(msa, wx, wy, wy+1) + d(msa, tx, ty, tx-1);

			if(e3i < e && eij != OO) {
				e = e3i;
				C_sfa[ij].set(ctype_3i, t);
			}
		}
		
	}
	C[ij] = e;
}


void RNAConsT::fillFM(const int & i, const int & j) {
	//case FM.(0)
	int e = OO;
	int ij = pos(i,j);
	FM_sfa[ij].ctype = ctype_0;
	
	vector<stem> vst = stl.getsubstems_within(i, j);
	FOR(k,0,sz(vst)) {
		stem t = vst[k];
		int tx = t.outer.lp; int ty = t.outer.rp;
		assert(tx>=i && ty<=j);
		int delta = C[pos(tx,ty)] + TP(msa, tx, ty) + N*TE.Mphi;
		//case FM.(1)
		int e1 = delta;
		if(e1 < e) {
			e = e1;
			FM_sfa[ij].set(ctype_1, t);
			FM_LS[ij] = t;
			FM_RS[ij] = t;
		}
		
		//case FM.(2)
		int eij = getE(etype_M1, ty+1, j);
		int e2 = delta + eij;
		if(e2 < e && eij != OO) {
			e = e2;
			FM_sfa[ij].set(ctype_2, t);
			FM_LS[ij] = t;
			FM_RS[ij] = M1_RS[pos(ty+1,j)];

		}
		
		//case FM.(3)
		stem w = getStem(M1_RS, ty+2, j);
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_M1, ty+2, j);
			int e3 = delta+ eij 
			         + min(d(msa, tx, ty, ty+1), d(msa, wx, wy, wx-1));
			if(e3 < e && eij != OO) {
				e = e3;
				FM_sfa[ij].set(ctype_3, t);
				FM_LS[ij] = t;
				FM_RS[ij] = M1_RS[pos(ty+2,j)];
			}
		}
		
		//case FM.(4)
		w = getStem(FM_RS, ty+3, j);
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_FM, ty+3, j);
			int e4 = delta + eij + 
			         d(msa, tx, ty, ty+1) + d(msa, wx, wy, wx-1);
			if(e4 < e && eij != OO) {
				e = e4;
				FM_sfa[ij].set(ctype_4, t);
				FM_LS[ij] = t;
				FM_RS[ij] = FM_RS[pos(ty+3,j)];
			}
		}
	}
	FM[ij] = e;
}
void RNAConsT::fillM1(const int & i, const int & j) {
	//case M1.(0)
	int e = OO;
	int ij = pos(i,j);
	M1_sfa[ij].ctype = ctype_0;
	vector<stem> vst = stl.getsubstems_startfrom_within(i, j);
	FOR(k,0,sz(vst)) {
		stem t = vst[k];
		int tx = t.outer.lp; int ty = t.outer.rp;
		assert(tx==i && ty<=j);
		int delta  = C[pos(tx,ty)] + N*TE.Mphi + TP(msa, tx, ty);
		//case M1.(1)
		int e1 = delta;
		if(e1 < e) {
			e = e1;
			M1_sfa[ij].set(ctype_1, t);
			M1_LS[ij] = t;
			M1_RS[ij] = t;
		}
		
		//case M1.(2)
		int eij = getE(etype_M1, ty+1, j);
		int e2 = delta + eij;
		if(e2 < e && eij != OO) {
			e = e2;
			M1_sfa[ij].set(ctype_2, t);
			M1_LS[ij] = t;
			M1_RS[ij] = M1_RS[pos(ty+1,j)];
		}
		
		//case M1.(3)
		stem w = getStem(M1_LS, ty+2, j);//M1_LS[ty+2][j];
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_M1, ty+2, j);
			int e3 = delta + eij
				      + min(d(msa, tx, ty, ty+1), d(msa, wx, wy, wx-1));
			if(e3 < e && eij != OO) {
				e = e3;
				M1_sfa[ij].set(ctype_3, t);
				M1_LS[ij] = t;
				M1_RS[ij] = M1_RS[pos(ty+2,j)];
			}
		}
		
		//case M1.(4)
		w = getStem(FM_LS, ty+3, j);
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_FM, ty+3, j);
			int e4 = delta + eij
			         + d(msa, tx, ty, ty+1) + d(msa, wx, wy, wx-1);
			if(e4 < e && eij != OO) {
				e = e4; 
				M1_sfa[ij].set(ctype_4, t);
				M1_LS[ij] = t;
				M1_RS[ij] = FM_RS[pos(ty+3,j)];
			}
		}
	}
	M1[ij] = e; 
}

void RNAConsT::fillM2(const int & i, const int & j) {
	//case M2.(0);
	int e = OO;
	int ij = pos(i,j);
	M2_sfa[ij].ctype = ctype_0;
	//M2_LS[i][j] remains to be (-1,-1,-1,-1)
	//M2_RS[i][j] remains to be (-1,-1,-1,-1)

	vector<stem> vst = stl.getsubstems_endat_within(i,j);
	FOR(k, 0, sz(vst)) {
		stem t = vst[k];
		int tx = t.outer.lp; int ty = t.outer.rp;
		assert(tx>=i && ty==j);
		int delta = C[pos(tx,ty)] + N*TE.Mphi + TP(msa, tx, ty);
		//case M2.(1)
		int e1 = delta;
		if(e1 < e) {
			e = e1; 
			M2_sfa[ij].set(ctype_1, t);
			M2_LS[ij] = t;
			M2_RS[ij] = t;
		}
		
		//case M2.(2)
		int eij = getE(etype_M2, i, tx-1);
		int e2 = delta + eij;
		if(e1 < e && eij != OO) {
			e = e1; 
			M2_sfa[ij].set(ctype_2, t);
			M2_LS[ij] = M2_LS[pos(i,tx-1)];
			M2_RS[ij] = t;
		}
		
		//case M2.(3)
		stem w = getStem(M2_RS, i, tx-2);
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_M2, i, tx-2);
			int e3 = delta + eij
			         + min(d(msa, wx, wy, wy+1), d(msa, tx, ty, tx-1));
			if(e3 < e && eij != OO) {
				M2_sfa[ij].set(ctype_3, t);
				M2_LS[ij] = M2_LS[pos(i,tx-2)];
				M2_RS[ij] = t;
			}
		}
		
		//case M2.(4)
		w = getStem(FM_LS, i, tx-3);
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_FM, i, tx-3);
			int e4 = delta +  eij 
			         + d(msa, wx, wy, wy+1) + d(msa, tx, ty, tx-1);
			if(e4 < e && eij != OO) {
				e = e4;
				M2_sfa[ij].set(ctype_4, t);
				M2_LS[ij] = FM_LS[pos(i,tx-3)];
				M2_RS[ij] = t;
			}
		}
	}
	M2[ij] = e;
}

void RNAConsT::fillM4(const int & i, const int & j) {
	//case M4.(0)
	int e = OO;
	int ij = pos(i,j);
	M4_sfa[ij].ctype = ctype_0;
	//M4_LS[i][j] remains to be (-1,-1,-1,-1)
	//M4_RS[i][j] remains to be (-1,-1,-1,-1)
	
	vector<stem> vst = stl.getsubstems_startfrom_within(i, j);
	FOR(k,0,sz(vst)) {
		stem t = vst[k];
		int tx = t.outer.lp; int ty = t.outer.rp;
		assert(tx==i && ty<=j);
		int delta = C[pos(tx,ty)] + N*TE.Mphi + TP(msa, tx, ty);
		int e1 = delta;
		//case M4.(1)
		if(e1 < e && ty==j) {
			e = e1;
			M4_sfa[ij].set(ctype_1, t);
			M4_LS[ij] = t;
			M4_RS[ij] = t;
		}
		
		//case M4.(2)
		int eij = getE(etype_M4, ty+1, j);
		int e2 = delta + eij;
		if(e2 < e && eij != OO) {
			e = e2;
			M4_sfa[ij].set(ctype_2, t);
			M4_LS[ij] = t;
			M4_RS[ij] = M4_RS[pos(ty+1,j)];
		}
		
		//case M4.(3)
		stem w = getStem(M4_LS, ty+2, j);
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_M4, ty+2, j);
			int e3 = delta + min(d(msa, tx, ty, ty+1), d(msa, wx, wy, wx-1)) + eij;
			if(e3 < e && eij != OO) {
				e = e3;
				M4_sfa[ij].set(ctype_3, t);
				M4_LS[ij] = t;
				M4_RS[ij] = M4_RS[pos(ty+2,j)];
			}
		}
				
		//case M4.(4)
		w = getStem(M2_LS, ty+3, j);
		if(w.valid()) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			eij = getE(etype_M2, ty+3, j);
			int e4 = delta + d(msa, tx, ty, ty+1) + d(msa, wx, wy, wx-1) + eij;
			if(e4 < e && eij != OO) {
				e = e4;
				M4_sfa[ij].set(ctype_4, t);
				M4_LS[ij] = t;
				M4_RS[ij] = M2_RS[pos(ty+3,j)];
			}
		}
	}
	M4[ij] = e;
}

void RNAConsT::fillF2n2d(const int & j) {
	//case (0)
	int ef2n = OO; int ef2d = OO;
	F2n_sfa[j].ctype = F2d_sfa[j].ctype = ctype_0;
	
	vector<stem> vst = stl.getsubstems_endat(j);
	FOR(k, 0, sz(vst)) {
		stem s = vst[k];
		int sx = s.outer.lp; int sy = s.outer.rp;
		int delta = C[pos(sx,sy)] + TP(msa, sx, sy);
		int d0 = d(msa,sx,sy,sy+1);

		stem t = getStem(F2n_S, sx-1);
		int ej = getE(etype_F2n, sx-1);
		//case (1)
		if(t.valid() && ej!=OO) {
			int ef2n1 = delta + ej;
			int ef2d1 = ef2n1 + d0;
			if(ef2n1 < ef2n) {
				ef2n = ef2n1;
				F2n_S[j] = s;
				F2n_sfa[j].set(ctype_1, s);
			}
			if(ef2d1 < ef2d) {
				ef2d = ef2d1;
				F2d_S[j] = s;
				F2d_sfa[j].set(ctype_1, s);
			}
		}
		
		//case (2)
		stem w = getStem(F2d_S, sx-2);
		ej = getE(etype_F2d, sx-2);
		if(w.valid() && ej!=OO) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int ef2n2 = ej + delta - d(msa,wx,wy,wy+1) + 
			            min(d(msa,wx,wy,wy+1), d(msa,sx,sy,sx-1));
			int ef2d2 = ef2n2 + d0;
			if(ef2n2 < ef2n) {
				ef2n = ef2n2;
				F2n_S[j] = s;
				F2n_sfa[j].set(ctype_2, s);
			}
			if(ef2d2 < ef2d) {
				ef2d = ef2d2;
				F2d_S[j] = s;
				F2d_sfa[j].set(ctype_2, s);
			}
		}
		
		//case (3)
		stem v = getStem(Fd_S, sx-3);
		ej = getE(etype_Fd, sx-3);
		if(v.valid() && ej!=OO) {
			int ef2n3 = delta + ej + d(msa,sx,sy,sx-1);
			int ef2d3 = ef2n3 + d0;
			if(ef2n3 < ef2n) {
				ef2n = ef2n3;
				F2n_S[j] = s; 
				F2n_sfa[j].set(ctype_3, s);
			}
			if(ef2d3 < ef2d) {
				ef2d = ef2d3;
				F2d_S[j] = s;
				F2d_sfa[j].set(ctype_3, s);
			}
		}
		//case (4)
		if(sx==0) {
			//case (4.a)
			int ef2n4a = delta;
			int ef2d4a = ef2n4a + d0;
			if(ef2n4a < ef2n) {
				ef2n = ef2n4a;
				F2n_S[j] = s;
				F2n_sfa[j].set(ctype_4a, s);
			}
			if(ef2d4a < ef2d) {
				ef2d = ef2d4a;
				F2d_S[j] = s;
				F2d_sfa[j].set(ctype_4a,s);
			}
		} else if(sx>=1) {
			//case (4.b)
			int ef2n4b = delta + d(msa,sx,sy,sx-1);
			int ef2d4b = ef2n4b + d0;
			if(ef2n4b < ef2n) {
				ef2n = ef2n4b;
				F2n_S[j] = s;
				F2n_sfa[j].set(ctype_4b, s);
			}
			if(ef2d4b < ef2d) {
				ef2d = ef2d4b;
				F2d_S[j] = s;
				F2d_sfa[j].set(ctype_4b,s);
			}
		}
	}
	
	F2n[j] = ef2n;
	F2d[j] = ef2d;
}

void RNAConsT::fillFnFd(const int & j) {
	//compute Fn[j] and Fd[j]
	//case FnFd.(0)
	if(j==0) {
		Fn[j] = Fd[j] = 0;
		Fn_sfa[j].ctype = Fd_sfa[j].ctype = ctype_0;
#ifdef DEBUG	
		cout<<"Fn["<<j<<"]="<<Fn[j]<<endl;
		cout<<"Fd["<<j<<"]="<<Fd[j]<<endl;
#endif
		return;
	}
	
	stem w = getStem(Fn_S, j-1);
	//case FnFd.(1)
	int efn = Fn[j-1];
	int efd = Fd[j-1];
	Fn_S[j] = Fn_S[j-1];
	Fd_S[j] = Fd_S[j-1];
	Fn_sfa[j].ctype = Fd_sfa[j].ctype = ctype_1;
		
	vector<stem> vst = stl.getsubstems_endat(j);
	FOR(k,0,sz(vst)) {
		stem s = vst[k];
		int sx = s.outer.lp; int sy = s.outer.rp;
		int delta = C[pos(sx,sy)] + TP(msa, sx, sy);
		int d0 = d(msa, sx,sy,sy+1);
		
		//case FnFd.(2)
		stem t = getStem(F2n_S, sx-1);
		int ej = getE(etype_F2n, sx-1);
		if(t.valid() && ej != OO) {
			int tx = t.outer.lp; int ty = t.outer.rp;
			int efn2 = ej + delta;
			int efd2 = efn2 + d0;
			if(efn2 < efn) {
				efn = efn2;
				Fn_S[j] = s;
				Fn_sfa[j].set(ctype_2, s);
			}
			if(efd2 < efd) {
				efd = efd2;
				Fd_S[j] = s;
				Fd_sfa[j].set(ctype_2, s);
			}
		}
		
		//case FnFd.(3)
		stem w = getStem(F2d_S, sx-2);
		ej = getE(etype_F2d, sx-2);
		if(w.valid() && ej != OO) {
			int wx = w.outer.lp; int wy = w.outer.rp;
			int efn3 = ej + delta - d(msa,wx,wy,wy+1) + 
			           min(d(msa,wx,wy,wy+1), d(msa,sx,sy,sx-1));
			int efd3 = efn3 +  d0;
			if(efn3 < efn) {
				efn = efn3;
				Fn_S[j] = s;
				Fn_sfa[j].set(ctype_3, s);
			}
			if(efd3 < efd) {
				efd = efd3;
				Fd_S[j] = s;
				Fd_sfa[j].set(ctype_3, s);
			}
		}
		
		//case FnFd.(4)
		stem v = getStem(Fd_S, sx-3);
		ej = getE(etype_Fd, sx-3);
		if(v.valid() && ej!=OO) {
			int efn4 = ej + delta + d(msa,sx,sy,sx-1);
			int efd4 = efn4 + d0;
			if(efn4 < efn) {
				efn = efn4;
				Fn_S[j] = s;
				Fn_sfa[j].set(ctype_4, s);
			}
			if(efd4 < efd) {
				efd = efd4;
				Fd_S[j] = s;
				Fd_sfa[j].set(ctype_4, s);
			}
		}
		
		//case FnFd.(5)
		if(sx==0) {
			//case FnFd.(5.a)
			int efn5a = delta;
			int efd5a = efn5a + d0;
			if(efn5a < efn) {
				efn = efn5a;
				Fn_S[j] = s;
				Fn_sfa[j].set(ctype_5a, s);
			}
			if(efd5a < efd) {
				efd = efd5a;
				Fd_S[j] = s;
				Fd_sfa[j].set(ctype_5a, s);
			}
		} else if(sx>=1) {
			//case FnFd.(5.b)
			int efn5b = delta + d(msa,sx,sy,sx-1);
			int efd5b = efn5b + d0;
			if(efn5b < efn) {
				efn = efn5b;
				Fn_S[j] = s;
				Fn_sfa[j].set(ctype_5b, s);
			}
			if(efd5b < efd) {
				efd = efd5b;
				Fd_S[j] = s;
				Fd_sfa[j].set(ctype_5b, s);
			}
		}
	}
	
	Fn[j] = efn;
	Fd[j] = efd;
}

void RNAConsT::fillF(const int & j) {
	if(j==0) {
		F[j] = 0;
		F_sfa[j].ctype = ctype_0;
		return;
	}	
	int ef = OO;
	int efn = Fn[j];
	int efd = Fd[j];
	stem w = getStem(Fn_S, j);
	int wx = w.outer.lp; int wy = w.outer.rp;
	
	stem v = getStem(Fd_S, j);
	int vx = v.outer.lp; int vy = v.outer.rp;
	
	if(w.valid()) {
		if(wy==j) {
			//	efn += 0;
		} else {
			efn += d(msa,wx,wy,wy+1);
		}
	}
	if(v.valid()) {
		if(vy==j) {
			efd -= d(msa,vx,vy,vy+1);
		} else {
			// efd += 0;
		}
	}
	
	if(efn < efd) {
		F[j] = efn;
		F_S[j] = w;
		F_sfa[j].set(ctype_1, w);
	} else {
		F[j] = efd;
		F_S[j] = v;
		F_sfa[j].set(ctype_2, v);
	}
}

string RNAConsT::traceback(){
	vector<stem> vst = tracebackF(L-1);
	string ret = "";
	FOR(i,0,L) ret += '.';
	
	FOR(i,0,sz(vst)) {
		stem s = vst[i];
		assert(s.valid());
		FOR(j,i+1,sz(vst)) {
			stem t = vst[j];
			assert(!s.conflictwith(t));
		}
	}
	
	FOR(i,0,sz(vst)) {
		stem s = vst[i];
		FOR(j,s.outer.lp, s.inner.lp+1) {
			ret[j] = '(';
		}
		FOR(j,s.inner.rp, s.outer.rp+1) {
			ret[j] = ')';
		}
	}
	return ret;
}

vector<stem> RNAConsT::tracebackF(const int & j) {
	assert(j>=0 && j<L);
	CTYPE ctypej = F_sfa[j].ctype;
	stem s = F_sfa[j].st;
	int sx = s.outer.lp; int sy = s.outer.rp;

	vector<stem> ret; 	
	switch(ctypej) {
		case ctype_0: {
			break; 
		}
		case ctype_1: {
			ret = tracebackFn(j);
			break;
		}
		case ctype_2: {
			ret = tracebackFd(j);
			break;
		}
	}
	return ret;
}

vector<stem> RNAConsT::tracebackFn(const int & j) {
	assert(j>=0 && j<L);
	CTYPE ctypej = Fn_sfa[j].ctype;
	stem s = Fn_sfa[j].st;
	int sx = s.outer.lp; int sy = s.outer.rp;
	
	vector<stem> vstC;
	vector<stem> vstF;
	switch(ctypej) {
		case ctype_0: {
			//do nothing
			break;
		} case ctype_1: {
			vstF = tracebackFn(j-1);
			break;
		} case ctype_2: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackF2n(sx-1);
			break;
		} case ctype_3: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackF2d(sx-2);
			break;
		} case ctype_4: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackFd(sx-3);
			break;
		} case ctype_5a: {
		} case ctype_5b: {
			vstC = tracebackC(sx,sy);
			break;
		} default: {
			assert(false);
		}
	}
	vector<stem> ret;
	ret.insert(ret.begin(), vstC.begin(), vstC.end());
	ret.insert(ret.begin(), vstF.begin(), vstF.end());
	return ret;	
}
vector<stem> RNAConsT::tracebackFd(const int & j) {
	assert(j>=0 && j<L);
	CTYPE ctypej = Fd_sfa[j].ctype;
	stem s = Fd_sfa[j].st;
	int sx = s.outer.lp; int sy = s.outer.rp;
	
	vector<stem> vstC;
	vector<stem> vstF;
	
	switch(ctypej) {
		case ctype_0: {
			//do nothing
			break;
		} case ctype_1: {
			vstF = tracebackFd(j-1);
			break;
		} case ctype_2: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackF2n(sx-1);
			break;
		} case ctype_3: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackF2d(sx-2);
			break;
		} case ctype_4: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackFd(sx-3);
			break;
		} case ctype_5a: {
		} case ctype_5b: {
			vstC = tracebackC(sx,sy);
			break;
		} default: {
			assert(false);
		}
	}
	
	vector<stem> ret;
	ret.insert(ret.begin(), vstC.begin(), vstC.end());
	ret.insert(ret.begin(), vstF.begin(), vstF.end());
	return ret;
}

vector<stem> RNAConsT::tracebackF2n(const int & j) {
	assert(j>=0 && j<L);
	return tracebackF2nF2d(j, F2n_sfa[j]);
}

vector<stem> RNAConsT::tracebackF2d(const int & j) {
	assert(j>=0 && j<L);
	return tracebackF2nF2d(j, F2d_sfa[j]);
}

vector<stem> RNAConsT::tracebackF2nF2d(const int & j, StackFoldAuxiliary & sfa) {
	assert(j>=0 && j<L);
	CTYPE ctypej = sfa.ctype;
	stem s = sfa.st;
	int sx = s.outer.lp; int sy = s.outer.rp;
	
	vector<stem> vstC;
	vector<stem> vstF;
	
	switch(ctypej) {
		case ctype_0: {
			assert(false);
			break;
		} case ctype_1: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackF2n(sx-1);
			break;
		} case ctype_2: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackF2d(sx-2);
			break;
		} case ctype_3: {
			vstC = tracebackC(sx,sy);
			vstF = tracebackFd(sx-3);
			break;
		} case ctype_4a: {
		} case ctype_4b: {
			vstC = tracebackC(sx,sy);
			break;
		} default: {
			assert(false);
		}
	}
	
	vector<stem> ret;
	ret.insert(ret.begin(), vstC.begin(), vstC.end());
	ret.insert(ret.begin(), vstF.begin(), vstF.end());
	return ret;
}

vector<stem> RNAConsT::tracebackC(const int & i, const int & j) {
	assert(i>=0 && i<L && j>=0 && j<L && i<j);
	vector<stem> ret;
	vector<stem> vstC;
	vector<stem> vstM;
	int ij = pos(i,j);
	stem s0 = stl.inwhichST(i,j);
	assert(s0.valid() && i-s0.outer.lp<=cf.MAXOVERLAP && s0.inner.lp-i+1>=2);

	stem s = stl.getsubstem_startfrom_endat(i, j);
	assert(s.valid() && s.outer.lp==i && s.outer.rp==j);

	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	ret.PB(s);

	//get stem t
	CTYPE ctypeij = C_sfa[ij].ctype;
	stem t = C_sfa[ij].st;
	int tx = t.outer.lp; int ty = t.outer.rp;
	int tp = t.inner.lp; int tq = t.inner.rp;
	assert(!s.conflictwith(t));
	
	//tracebackC(t)
	if(ctypeij != ctype_1) {
		assert(t.valid());
		vstC = tracebackC(tx,ty);
	}	
	//get i1, i2, i3, j1, j2, j3
	int i1 = sp+1, i2 = sp+2, i3 = sp+3;
	int j1 = tx-1, j2 = tx-2, j3 = tx-3;
	switch(ctypeij) {
		case ctype_1: {
			//hairpin loop, do nothing
			break;
		} case ctype_2: {
			//interior loop or bulge formed between t and s,
			//tracebackC(tx,ty) has been executed before switch(ctypeij)
			break;
		} case ctype_3a: {
			vstM = tracebackM4(i1,j1);
			break;
		} case ctype_3b: {
			vstM = tracebackM4(i1,j2);
			break;
		} case ctype_3c: {
			vstM = tracebackM4(i1,j3);
			break;
		} case ctype_3d: {
			vstM = tracebackM4(i2,j1);
			break;
		} case ctype_3e: {
			vstM = tracebackM4(i2,j2);
			break;
		} case ctype_3f: {
			vstM = tracebackM4(i2,j3);
			break;
		} case ctype_3g: {
			vstM = tracebackM4(i3,j1);
			break;
		} case ctype_3h: {
			vstM = tracebackM4(i3,j2);
			break;
		} case ctype_3i: {
			vstM = tracebackFM(i3,j3);
			break;
		} case ctype_3p: {
			vstM = tracebackM2(i3,j1);//tracebackFM(i3,j3);
			break;
		} case ctype_3q: {
			vstM = tracebackM2(i3,j2);//tracebackFM(i3,j3);
			break;
		} case ctype_3r: {
			vstM = tracebackM1(i1,j3);//tracebackFM(i3,j3);
			break;
		} case ctype_3s: {
			vstM = tracebackM1(i2,j3);//tracebackFM(i3,j3);
			break;
		} default :{
			assert(false);
		}
	}
	ret.insert(ret.begin(), vstC.begin(), vstC.end());
	ret.insert(ret.begin(), vstM.begin(), vstM.end());
	return ret;
}

vector<stem> RNAConsT::tracebackFM(const int & i, const int & j) {
	assert(i>=0 && i<L && j>=0 && j<L && i<j);
	vector<stem> ret;
	vector<stem> vstC; 
	vector<stem> vstM;
	int ij = pos(i,j);
	CTYPE ctypeij = FM_sfa[ij].ctype;
	stem t = FM_sfa[ij].st;
	int tx = t.outer.lp; int ty = t.outer.rp;
	switch(ctypeij){
		case ctype_0: {
			assert(false);
			break; 
		} case ctype_1: {
			vstC = tracebackC(tx,ty);
			break; 
		} case ctype_2: {
			vstC = tracebackC(tx,ty);
			vstM = tracebackM1(ty+1,j);
			break;
		} case ctype_3: {
			vstC = tracebackC(tx,ty);
			vstM = tracebackM1(ty+2,j);
			break; 
		} case ctype_4: {
			vstC = tracebackC(tx,ty);
			vstM = tracebackFM(ty+3,j);
			break; 
		} default:{
			assert(false);
		}
	}
	ret.insert(ret.begin(), vstC.begin(), vstC.end());
	ret.insert(ret.begin(), vstM.begin(), vstM.end());
	return ret;
}

vector<stem> RNAConsT::tracebackM1(const int & i, const int & j) {
	assert(i>=0 && i<L && j>=0 && j<L && i<j);
	vector<stem> ret;
	int ij = pos(i,j);
	CTYPE ctypeij = M1_sfa[ij].ctype;
	stem t = M1_sfa[ij].st;
	int tx = t.outer.lp; int ty = t.outer.rp;
	vector<stem> vstC;
	vector<stem> vstM;
	
	switch(ctypeij){
		case ctype_0: {
			assert(false);
			break; 
		} case ctype_1: {
			vstC = tracebackC(tx,ty);
			break; 
		} case ctype_2: {
			vstC = tracebackC(tx,ty);
			vstM = tracebackM1(ty+1,j);
			break; 
		} case ctype_3: {
			vstC = tracebackC(tx,ty);
			vstM = tracebackM1(ty+2,j);
			break; 
		} case ctype_4: {
			vstC = tracebackC(tx,ty);
			vstM = tracebackFM(ty+3,j);
			break; 
		} default:{
			assert(false);
		}
	}
	ret.insert(ret.begin(), vstC.begin(), vstC.end());
	ret.insert(ret.begin(), vstM.begin(), vstM.end());
	return ret;
}

vector<stem> RNAConsT::tracebackM2(const int & i, const int & j) {
	assert(i>=0 && i<L && j>=0 && j<L && i<j);
	vector<stem> ret;
	int ij = pos(i, j);
	CTYPE ctypeij = M2_sfa[ij].ctype;
	stem t = M2_sfa[ij].st;
	int tx = t.outer.lp; int ty = t.outer.rp;
	assert(ty==j);
	vector<stem> vstC;
	vector<stem> vstM;
	switch(ctypeij){
		case ctype_0: {
			assert(false);
			break; 
		} case ctype_1: {
			vstC = tracebackC(tx, ty);
			break; 
		} case ctype_2: {
			vstC = tracebackC(tx, ty);
			vstM = tracebackM2(i,tx-1);
			break; 
		} case ctype_3: {
			vstC = tracebackC(tx, ty);
			vstM = tracebackM2(i,tx-2);
			break;
		} case ctype_4: {
			vstC = tracebackC(tx, ty);
			vstM = tracebackFM(i,tx-3);
			break; 
		} default:{
			assert(false);
		}
	}
	ret.insert(ret.begin(), vstC.begin(), vstC.end());
	ret.insert(ret.begin(), vstM.begin(), vstM.end());
	return ret;
}

vector<stem> RNAConsT::tracebackM4(const int & i, const int & j) {
	assert(i>=0 && i<L && j>=0 && j<L && i<j);
	vector<stem> ret;
	int ij = pos(i,j);
	CTYPE ctypeij = M4_sfa[ij].ctype;
	stem t = M4_sfa[ij].st;
	int tx = t.outer.lp; int ty = t.outer.rp;
	assert(t.valid() && tx==i);

	vector<stem> vstC;
	vector<stem> vstM;
	
	switch(ctypeij){
		case ctype_0: {
			assert(false);
			break; 
		} case ctype_1: {
			vstC = tracebackC(tx, ty);
			break; 
		} case ctype_2: {
			vstC = tracebackC(tx, ty);
			vstM = tracebackM4(ty+1,j);
			break; 
		} case ctype_3: {
			vstC = tracebackC(tx, ty);
			vstM = tracebackM4(ty+2,j);
			break; 
		} case ctype_4: {
			vstC = tracebackC(tx, ty);
			vstM = tracebackM2(ty+3,j);
			break; 
		} default:{
			assert(false);
		}
	}
	ret.insert(ret.begin(), vstC.begin(), vstC.end());
	ret.insert(ret.begin(), vstM.begin(), vstM.end());
	return ret;
}
int RNAConsT::getE(const ETYPE & etype, const int & i, const int & j) {
	int ret = OO;
	int ij = pos(i,j);
	switch(etype) {	
		case etype_F: {
			if(!(i<0 || i>=L)) ret = F[i];
			assert(j==-1); break;
		} case etype_Fn: {
			if(!(i<0 || i>=L)) ret = Fn[i];
			assert(j==-1); break;
		} case etype_Fd: {
			if(!(i<0 || i>=L)) ret = Fd[i];
			assert(j==-1); break;
		} case etype_F2n: {
			if(!(i<0 || i>=L)) ret = F2n[i];
			assert(j==-1); break;
		} case etype_F2d: {
			if(!(i<0 || i>=L)) ret = F2d[i];
			assert(j==-1); break;
		} case etype_C : {
			if(!(i<0 || i>=L || j<0 || j>=L || (i>j) || (j<i+4)))
				ret = C[ij];
			break;
		} case etype_FM: {
			if(!(i<0 || i>=L || j<0 || j>=L || (i>j)))
				ret = FM[ij];
			break;
		} case etype_M1: {
			if(!(i<0 || i>=L || j<0 || j>=L || (i>j)))
				ret = M1[ij];
			break;
		} case etype_M2: {
			if(!(i<0 || i>=L || j<0 || j>=L || (i>j)))
				ret = M2[ij];
			break;
		} case etype_M4: {
			if(!(i<0 || i>=L || j<0 || j>=L || (i>j)))
				ret = M4[ij];
			break;
		} default: {
			showETYPE(etype);
			assert(false);
		}
	}
	ret = min(OO, ret);
	return ret;
}

stem RNAConsT::getStem(const vector<stem> & vstem, const int & i, const int & j) {
	return getStem(vstem, pos(i,j));
}
	
stem RNAConsT::getStem(const vector<stem> & vstem, const int & j) {
	stem ret(-1,-1,-1,-1);
	if(j>=0 && j<=pos(L-1,L-1)) {
		ret = vstem[j];
	}
	return ret;
}

int RNAConsT::computeEnergy(stemlist & in_stl) {
	if(sz(in_stl)==0) return 0;
	int ret = 0;
	vector<int> vinside;
	//vinside[i] contains a 
	//0 - Hairpin ((((....))))
	//1 - Interior loop or bulge ((((..((((....))))..))))
	//n, n>=2 - Multiple loops ((((....((((....))))....((((....)))))....))))
	//where n is the number of branches

	vector<int> voutside;
	//0 - not enclosed by any stem
	//1 - enclosed by an interior loop
	//2 - enclosed by a multi loop

	vector<vector<int>> relations;
	
	FOR(i,0,sz(in_stl.vstems))	{
		vinside.PB(0);
		voutside.PB(0);
	}
//	cout<<"show stl"<<endl;
//	FOR(i,0,sz(in_stl.vstems)) {
//		in_stl.vstems[i].show(); cout<<endl;
//	}
//	cout<<"end show"<<endl;
	//relations[i][j] has the relationships between i and j
	//0  - no relationship
	//1  - sti enclose stj and there does not exist any other stem stk such that stk enclose stj and sti enclose stk
	//-1  - sti is enclosed by stj, and there does not exist any other stem stk such that stk sti is enclosed by stk and stk is enclosed by stj 
   //2  - sti enclose stj and there exist a stem stk such that ...
   //-2 - sti is enclosed by stj, and there exist a stem stk such that ...	
	FOR(i,0,sz(in_stl.vstems)) {
		vector<int> tmpvi;
		relations.PB(tmpvi);
		FOR(j,0,sz(in_stl.vstems)){
			relations[i].PB(0);
		}
	}
	
	FOR(i,0,sz(in_stl.vstems)) {
		stem sti = in_stl.vstems[i];
		FOR(j,0,sz(in_stl.vstems)) {
			stem stj = in_stl.vstems[j];
			if(i==j) {
				relations[i][j] = 0; 
				continue;
			}
			if(sti.enclose(stj)) {
				bool exist = false;
				FOR(k,0,sz(in_stl.vstems)){
					if(k==i || k==j) continue;
					stem stk = in_stl.vstems[k];
					if(sti.enclose(stk) && stk.enclose(stj)){
						exist = true;
						break;
					}
				}
				if(exist) relations[i][j] = 2;
				else relations[i][j] = 1;
			} else if(stj.enclose(sti)){
				bool exist = false;
				FOR(k,0,sz(in_stl.vstems)){
					if(k==i || k==j) continue;
					stem stk = in_stl.vstems[k];
					if(stj.enclose(stk) && stk.enclose(sti)){
						exist = true;
						break;
					}
				}
				if(exist) relations[i][j] = -2;
				else relations[i][j] = -1;
			} else {
				relations[i][j] = 0;
			}
		}
	}

	FOR(i,0,sz(in_stl.vstems)) {
		stem sti = in_stl.vstems[i];
		FOR(j,0,sz(in_stl.vstems)) {
			if(relations[i][j] == 1) {
				vinside[i]++;
			}
		}
	}
	
	FOR(i,0,sz(in_stl.vstems)) {
		stem sti = in_stl.vstems[i];
		FOR(j,0,sz(in_stl.vstems)) {
			if(relations[i][j] == -1) {
				if(vinside[j]==0) assert(false);
				else if(vinside[j]==1) {
					voutside[i] = 1;
				} else if(vinside[j]>=2) {
					voutside[i] = 2;
				}
			}
		}
	}
	
	vector<int> vstindx;
	vstindx.PB(0);
	int prei = -1;
	while(sz(vstindx)>0){
		int i = vstindx[0];
		assert(i>prei);
		prei = i;
		stem st = in_stl.vstems[i];

		vstindx.erase(vstindx.begin());
		
		if(vinside[i] == 0) {
			ret += computeH(msa, st);
		} else if (vinside[i] == 1){
			for(int k=sz(in_stl.vstems)-1; k>=0; k--) {
				if(relations[i][k] == 1) {
					stem stk = in_stl.vstems[k];
					ret += computeS(msa, st)+computephi(msa, st, stk);
					vstindx.insert(vstindx.begin(),k);
					break;
				}
			}
		} else if (vinside[i] >= 2){
			ret += computeS(msa, st) + N*(TE.Mphi + TE.Mc);
			
			for(int k=sz(in_stl.vstems)-1; k>=0; k--) {
				if(relations[i][k] == 1) {
					ret += N*TE.Mphi;
					vstindx.insert(vstindx.begin(), k);
				}
			}
		} else assert(false);
		
		if(sz(vstindx)==0 && i<sz(in_stl.vstems)-1) {
			vstindx.PB(i+1);
		}
	}
	FOR(i,0,sz(in_stl.vstems)) {
		stem st = in_stl.vstems[i];
		int x = st.outer.lp; int y = st.outer.rp;
		int p = st.inner.lp; int q = st.inner.rp;
		
		if(voutside[i]==1) {
			//inside an interior loop, no terminal AU penalty for x,y
		} else {
			ret += TP(msa, x, y);
		}
		
		if(vinside[i]==1 || vinside[i]==0) {
			//enclose an interior loop, no terminal AU penalty for p,q
			//a hairpin loop, no temrinalAU penalty for p,q
		} else {
			ret += TP(msa, p, q);
		}
		
		int k = x-1; int h = x-2;
		int a = y+1; int b = y+2;
		int c = p+1; int m = p+2;
		int e = q-1; int f = q-2;
		int k1, h1, a1, b1, c1, m1, e1, f1;
		k1 = h1 = a1 = b1 = c1 = m1 = e1 = f1 = -1;
		FOR(j,0,sz(in_stl.vstems)) {
			if(i==j) continue;
			stem stj = in_stl.vstems[j];
			if(stj.pairwithwhich(k)!=-1) k1 = stj.pairwithwhich(k);
			if(stj.pairwithwhich(h)!=-1) h1 = stj.pairwithwhich(h);
			if(stj.pairwithwhich(a)!=-1) a1 = stj.pairwithwhich(a);
			if(stj.pairwithwhich(b)!=-1) b1 = stj.pairwithwhich(b);
			if(stj.pairwithwhich(c)!=-1) c1 = stj.pairwithwhich(c);
			if(stj.pairwithwhich(m)!=-1) m1 = stj.pairwithwhich(m);
			if(stj.pairwithwhich(e)!=-1) e1 = stj.pairwithwhich(e);
			if(stj.pairwithwhich(f)!=-1) f1 = stj.pairwithwhich(f);
		}
		
		if(k1!=-1) {
			//seq[k] is pairing with seq[k1], don't add d(x,y,k)
		} else if(k1==-1 && h1!=-1) {
			if(voutside[i]==1) {
			//interior loop, no dangle energy contributions
			} else {
				if(h1>h) {
					//min(d(x,y,k), d(h1,h,k)) should have been included when processing a stem with inner=(h,h1)
				} else {
					//min(d(x,y,k), d(h1,h,k)) should have been included when processing a stem enclosed by (h1,h)
				}
			}
		} else if(k1==-1 && h1==-1) {
			if(voutside[i]==1) {
			//interior loop, no dangle energy contributions
			} else {
				ret += d(msa, x, y, k);
			}
		}
		if(a1!=-1) {
			//seq[a] is pairing with seq[a1], don't add d(x,y,a)
		} else if(a1==-1 && b1!=-1) {
			if(voutside[i]==1) {
				//interior loop, no dangle energy contributions
			} else {
				if(b1>b) {
					ret += min(d(msa, x, y, a), d(msa, b, b1, a));
				} else {
					ret += min(d(msa, x, y, a), d(msa, b1, b, a));
				}
			}
		} else if(a1==-1 && b1==-1) {
			if(voutside[i]==1) {
				//interior loop, no dangle energy contributions
			} else {
				ret += d(msa, x, y, a);
			}
		}
		
		if(c1!=-1) {
			//dont add d(p,q,c)
		} else if(c1==-1 && m1!=-1) {
			if(vinside[i]==0) {//Hairpin
				assert(false);
			} else if(vinside[i]==1) {//interior or bulge
			} else if(vinside[i]>=2) {//multiloop
				ret += min(d(msa, p, q, c), d(msa, m, m1, c));
			}
		} else if(c1==-1 && m1==-1) {
			if(vinside[i]==0) {//Hairpin
			} else if(vinside[i]==1) {//interior or bulge
			} else if(vinside[i]>=2) {//multiloop
				ret += d(msa, p, q, c);
			}
		}

		if(e1!=-1){
		} else if(e1==-1 && f1!=-1) {
			if(vinside[i]==0) {//Hairpin
				assert(false);
			} else if(vinside[i]==1) {//interior or bulge
			} else if(vinside[i]>=2) {//multiloop
				//min(d(p,q,e), d(f1, f, e)) should have been included when processing a stem enclosed by (f1, f);
			}
		} else if(e1==-1 && f1==-1) {
			if(vinside[i]==0) {//Hairpin
			} else if(vinside[i]==1) {//interior or bulge
			} else if(vinside[i]>=2) {//multiloop
				ret += d(msa, p, q, e);
			}
		}
	}
	return ret;
}

int RNAConsT::computeScov(stemlist & in_stl) {
	int ret = 0;
	FOR(i,0, sz(in_stl.vstems)) {
		stem sti = in_stl.vstems[i];
		assert(sti.valid());
		FOR(j,sti.outer.lp,sti.inner.lp+1) {
			ret += c[j][sti.outer.lp+sti.outer.rp-j];
		}
	}
	return ret;
}


int RNAConsT::getsegment_esum(seglistT & in_sgl) {
	int ret = 0;
	FOR(k,0,sz(in_sgl)) {
		segmentT sg = in_sgl[k];
		int i = sg.lp; int j = sg.rp; ETYPE type = sg.etype;
		int ij = pos(i, j);
		switch(type) {
			case  etype_F: {ret += F[j]; assert(i==0); break;}
			case  etype_C: {ret += C[ij]; break;}
			case  etype_FM: {ret += FM[ij]; break;}
		}
	}
	return ret;
}


vector<RNA> RNAConsT::generate_subopt(ofstream & fout, int & e_ub, bool & reqLocOpt) {
	vector<RNA> ret;
	int szstl = sz(stl);	
	vector<PartStackStruct> R;
	
	PartStackStruct phi;
	segmentT sg_n(0, L-1, etype_F);
	phi.sgl.PB(sg_n);
	phi.elphi = 0;
	//phi <- ({[0..L-1], F}, {}, 0)

	R.PB(phi); //R.PB(phi);
	int indx = 0;
	int num = 0;
	while(sz(R)!=0) {
		indx++;
		bool happen = false;
		PartStackStruct phi_0 = R[0];
		//(theta, P, e, log) <- phi_0
		seglistT theta = phi_0.sgl;
		stemlist P = phi_0.stl;
		int e = phi_0.elphi;
		string log = phi_0.log;
		R.erase(R.begin());

		if(sz(theta)==0) {
			//print the secondary structure of phi
			RStack rst(P.vstems, L);
			int eprim = computeEnergy(P);
			int ecov = computeScov(P);
			if(eprim <= e_ub) {
				num++;
				if(num%500==0) {
					cout<<"The algorithm has generated "<<num<<" local optimal stack configurations"<<endl;
				}
				ret.PB(RNA(rst.tostring(), eprim, ecov));
			}
			continue;
		}
		//assert sz(theta) != 0
		
		segmentT sg_0 = theta[sz(theta)-1]; theta.vsegs.erase(theta.vsegs.end()-1);
		ETYPE type_0 = sg_0.etype;

		PartStackStruct pprim(theta, P, e, log);
		switch(type_0) {
			case etype_F: {happen = suboptF(R, sg_0, pprim, e_ub, reqLocOpt); break;}
			case etype_C: {happen = suboptC(R, sg_0, pprim, e_ub, reqLocOpt); break;}
			case etype_FM:{happen = suboptFM(R, sg_0, pprim, e_ub, reqLocOpt);break;}
			default : {cerr<<"ETYPE="<<type_0<<" is unknown."<<endl; exit(0);}
		}
	}
	return ret;
}

bool RNAConsT::suboptF(vector<PartStackStruct> & R, 
     const segmentT & in_sg, const PartStackStruct & in_phi, const int & e_ub, 
     const bool & reqLocOpt) {
   PartStackStruct tmp = in_phi;
	bool happen = false;
	int i = in_sg.lp;
	int j = in_sg.rp;
	assert(in_sg.etype == etype_F);
	assert(i==0);
	
	PartStackStruct pprim0 (in_phi);
	stem w = pprim0.stl.findRS(segment(0, j));
	assert(!w.valid() || (w.valid() && w.outer.lp > j));
	int wx = w.outer.lp; int wy = w.outer.rp;
	

	vector<stem> vst;
	if(w.valid() && w.outer.lp == j+1)
		vst = stl.getsubstems_endat(j);
	else 
		vst = stl.getstems_endat(j);
	FOR(k, 0, sz(vst)) {
		stem t = vst[k];
		int tx = t.outer.lp; int ty = t.outer.rp;
		int tp = t.inner.lp; int tq = t.inner.rp;
			
		PartStackStruct pprim (in_phi);
		int delta = 0;
		if(w.valid()) {
			if(ty==wx-1) 
				delta += 0;
			else if(ty==wx-2)
				delta += min(d(msa,tx,ty,ty+1), d(msa,wx,wy,wx-1));
			else if(ty<=wx-3)
				delta += d(msa,tx,ty,ty+1) + d(msa,wx,wy,wx-1);
		} else {
			if(j==L-1) {
				delta += 0;
			} else if(j<=L-2) {
				delta += d(msa,tx,ty,ty+1);
			}
		}
		
		int delta2 = 0;
		if(tx==0)
			delta2 += 0;
		else if(tx>=1) 
			delta2 += d(msa,tx,ty,tx-1);
		
		segmentT sg_t(tx,ty, etype_C);
		pprim.sgl.PB(sg_t);
		pprim.elphi += delta+delta2+TP(msa, tx, ty);
		int esum = getsegment_esum(pprim.sgl);
		int emin = esum+pprim.elphi;

		if((!reqLocOpt || (reqLocOpt && pprim.isLocOpt(stl))) && emin <= e_ub) {
			stringstream ss;
			ss<<"elphi="<<pprim.elphi<<"+dR("<<delta<<")+dL("<<delta2<<")+AU("<<TP(msa, tx, ty)<<")=";
			string newlog=ss.str();
			pprim.log += newlog;
			R.PB(pprim);
			happen = true;
		}
		if(tx<2*cf.MINSTACKLEN+cf.MINLOOPLEN-1-cf.MAXOVERLAP) {
			continue;
		}
	
		PartStackStruct pprim2(in_phi);
		segmentT sg_f(0, tx-1, etype_F);
		pprim2.sgl.PB(sg_f);
		pprim2.sgl.PB(sg_t);
		pprim2.elphi += delta + TP(msa, tx, ty);
		esum = getsegment_esum(pprim2.sgl);
		//-170*N is the minimum dangling energy to the right side of sg_f
			//d(msa, tx, ty, tx-1) is the minimum dangling energy to the left side of sg_t
		emin = esum + pprim2.elphi + (-170)*N + d(msa, tx, ty, tx-1);
		
		if((!reqLocOpt || (reqLocOpt && pprim2.isLocOpt(stl))) && emin <= e_ub) {
			stringstream ss;
			ss<<"subopt F["<<i<<"]["<<j<<"], elphi="<<pprim2.elphi<<"=elphi+dR("<<delta<<")+AU("<<TP(msa, tx, ty)<<")"<<endl;
			string newlog=ss.str();
			pprim2.log += newlog;
			R.PB(pprim2);
			happen = true;
		}
	}
		

	if (j<2*cf.MINSTACKLEN+cf.MINLOOPLEN-1-cf.MAXOVERLAP) {
	} else {
		segmentT sg_f(0,j-1, etype_F);
		PartStackStruct pprim3 (in_phi);
		pprim3.sgl.PB(sg_f); //pprim3.sgt.PB(etype_F);
		int esum = getsegment_esum(pprim3.sgl);//, pprim3.sgt);
		int emin = esum + pprim3.elphi + (-170*2)*N;
		if((!reqLocOpt || (reqLocOpt && pprim3.isLocOpt(stl))) && emin <= e_ub) 
		{
			stringstream ss;
			ss<<"subopt F["<<i<<"]["<<j<<"], elphi="<<pprim3.elphi<<endl;
			string newlog=ss.str();
			pprim3.log += newlog;
			R.PB(pprim3);
			happen = true;
		}
	} 
	return happen;
}

bool RNAConsT::suboptFM(vector<PartStackStruct> & R, 
     const segmentT & in_sg, const PartStackStruct & in_phi, const int & e_ub, 
     const bool & reqLocOpt) {
	PartStackStruct tmp = in_phi;
#ifdef DEBUG
	cout<<"suboptFM("<<in_sg.lp<<","<<in_sg.rp<<", "; tmp.show(); cout<<endl;
#endif
	bool happen = false;
	int i = in_sg.lp;
	int j = in_sg.rp;
	assert(in_sg.etype == etype_FM);
	
	PartStackStruct pprim0 (in_phi);
	stem v = pprim0.stl.findLS(in_sg);
	stem w = pprim0.stl.findRS(in_sg);
	assert(v.valid() && w.valid() && v.enclose(w));
	int vx = v.outer.lp; int vy = v.outer.rp;
	int vp = v.inner.lp; int vq = v.inner.rp;
	int wx = w.outer.lp; int wy = w.outer.rp;
	assert(vp == i-1 &&  j < vq);
	
	vector<stem> vst;
	if (wx==j+1) {
		vst = stl.getsubstems_between(v, w);
	} else {
		vst = stl.getstems_within(i,j);
	}

	FOR(k,0,sz(vst)) {
		stem t = vst[k];
		int tx = t.outer.lp; int ty = t.outer.rp;
		int tp = t.inner.lp; int tq = t.inner.rp;
		
		PartStackStruct pprim (in_phi);
		int delta = 0;
		if(ty==wx-1) 
			delta += 0;
		else if(ty==wx-2)
			delta += min(d(msa,tx,ty,ty+1), d(msa,wx,wy,wx-1));
		else if(ty<=wx-3)
			delta += d(msa,tx,ty,ty+1) + d(msa,wx,wy,wx-1);
		int delta2 = 0;
		if(tx==vp+1)
			delta2 += 0;
		else if(tx==vp+2)
			delta2 += min(d(msa,vp,vq,vp+1), d(msa,tx,ty,tx-1));
		else if(tx>=vp+3)
			delta2 += d(msa,vp,vq,vp+1) + d(msa,tx,ty,tx-1);
				
		segmentT sg_t(tx,ty, etype_C);
		pprim.sgl.PB(sg_t);
		pprim.elphi += delta + TP(msa,tx,ty) + delta2 + 
		               N*TE.Mphi;
		int esum = getsegment_esum(pprim.sgl);//, pprim.sgt);
		int emin = esum + pprim.elphi;

		if((!reqLocOpt || (reqLocOpt && pprim.isLocOpt(stl))) && emin <= e_ub) {
			stringstream ss;
			ss<<"subopt FM["<<i<<"]["<<j<<"], elphi="<<pprim.elphi<<"=elphi+dR("<<delta<<")+dL("<<delta2<<")+AU("<<TP(msa,tx,ty)<<")+Mphi(40)"<<endl;
			string newlog=ss.str();
			pprim.log += newlog;		
			R.PB(pprim);

			happen = true;
		}
		
		if(tx-vp<2*cf.MINSTACKLEN+cf.MINLOOPLEN+1-2*cf.MAXOVERLAP)	{
			continue;
		}

		PartStackStruct pprim2 (in_phi);
		segmentT sg_fm(i, tx-1, etype_FM);
		pprim2.sgl.PB(sg_fm); //pprim2.sgt.PB(etype_FM);
		pprim2.sgl.PB(sg_t); //pprim2.sgt.PB( etype_C);
		pprim2.elphi += delta + TP(msa, tx, ty) + N*TE.Mphi;

		esum = getsegment_esum(pprim2.sgl);//, pprim2.sgt);
		emin = esum + pprim2.elphi + (-170)*4*N;
		if((!reqLocOpt || (reqLocOpt && pprim2.isLocOpt(stl))) && emin <= e_ub) 
		{
			stringstream ss;
			ss<<"subopt FM["<<i<<"]["<<j<<"], elphi="<<pprim2.elphi<<"=elphi+dR("<<delta<<")+AU("<<TP(msa, tx, ty)<<")"<<endl;
			string newlog=ss.str();
			pprim2.log += newlog;
			R.PB(pprim2);
#ifdef DEBUG
			cout<<"Push back FM.2, "; pprim2.show();
#endif
			happen = true;
		}
	}
	return happen;
}


bool RNAConsT::suboptC(vector<PartStackStruct> & R, 
     const segmentT & in_sg, const PartStackStruct & in_phi, const int & e_ub, 
     const bool & reqLocOpt) {
   PartStackStruct tmp = in_phi;

	bool happen = false;
	int i = in_sg.lp;
	int j = in_sg.rp;
	assert(in_sg.etype == etype_C);
	
	stem s0 = stl.inwhichST(i,j);
	assert(s0.valid() && i-s0.outer.lp<=cf.MAXOVERLAP && s0.inner.lp-i+1>=2);
	
	stem s = stl.getsubstem_startfrom_endat(i,j);
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	assert(s.valid() && sx==i && sy==j);
	
	PartStackStruct pprim (in_phi);
	pprim.stl.add(s);
	int eH = computeH(msa, s);
	pprim.elphi += eH;
	int esum = getsegment_esum(pprim.sgl);//, pprim.sgt);
	int emin = esum + pprim.elphi;

	if((!reqLocOpt || (reqLocOpt && pprim.isLocOpt(stl))) && emin <= e_ub) {
		stringstream ss;
		ss<<"subopt C["<<i<<"]["<<j<<"], elphi="<<pprim.elphi<<"="<<in_phi.elphi<<"+H["<<sx<<","<<sy<<"]["<<sp<<","<<sq<<"]("<<eH<<")"<<endl;
		string newlog=ss.str();
		pprim.log += newlog;
		R.PB(pprim);
		happen = true;
	}
	
	vector<stem> vst = stl.getsubstems_enclosedby_or_inward_extend(s);
	FOR(k,0,sz(vst)) {
		stem t = vst[k];
		int tx = t.outer.lp;  int ty = t.outer.rp;
		int tp = t.inner.lp;  int tq = t.inner.rp;
		
		//interior loop
		PartStackStruct pprim (in_phi);
		segmentT sg_t(tx, ty, etype_C);
		pprim.sgl.PB(sg_t);
		pprim.stl.add(s);
		int eI = computeS(msa, s) + computephi(msa, s, t);
		pprim.elphi += eI;//S[indx_s*szstl+indx_t]+phi[indx_s*szstl+indx_t];
		
		esum = getsegment_esum(pprim.sgl);
		emin = esum + pprim.elphi;
		if((!reqLocOpt || (reqLocOpt && pprim.isLocOpt(stl))) && emin <= e_ub) {
			stringstream ss;
			ss<<"subopt C["<<i<<"]["<<j<<"], elphi="<<pprim.elphi<<"="<<in_phi.elphi<<"+S["<<sx<<","<<sy<<"~"<<sp<<","<<sq<<"]("<<computeS(msa, s)<<")+phi["<<sx<<","<<sy<<"~"<<sp<<","<<sq<<"]["<<tx<<","<<ty<<"~"<<tp<<","<<tq<<"]("<<computephi(msa,s,t)<<")"<<endl;
			string newlog=ss.str();
			pprim.log += newlog;
			R.PB(pprim);
			happen = true;
		}
		
		if(tx-(sp+1) < 2*cf.MINSTACKLEN+cf.MINLOOPLEN-2*cf.MAXOVERLAP) 
			continue;
		//multi-loop
		int delta = 0;
		if(ty==sq-1) {
			delta += 0;
		} else if(ty==sq-2) {
			delta += min(d(msa,tx,ty,ty+1), d(msa,sp,sq,sq-1));
		} else {
			delta += d(msa,tx,ty,ty+1) + d(msa,sp,sq,sq-1);
		}
		segmentT sg_fm(sp+1,tx-1, etype_FM);
		PartStackStruct pprim2(in_phi);

		pprim2.sgl.PB(sg_fm); //pprim2.sgt.PB(etype_FM);
		pprim2.sgl.PB(sg_t); //pprim2.sgt.PB(etype_C);
		pprim2.stl.add(s);
		pprim2.elphi += computeS(msa, s)+
		                N*(TE.Mc+2*TE.Mphi)+
		                TP(msa, sp, sq)+delta+
		                TP(msa, tx, ty);
      esum = getsegment_esum(pprim2.sgl);//, pprim2.sgt);
		emin = esum + pprim2.elphi + (-170*4)*N;

		if((!reqLocOpt || (reqLocOpt && pprim2.isLocOpt(stl))) && emin <= e_ub) {
			stringstream ss;
			ss<<"subopt C["<<i<<"]["<<j<<"], s=("<<sx<<","<<sp<<")~("<<sq<<","<<sy<<"), t=("<<tx<<","<<tp<<")~("<<tq<<","<<ty<<")"<<endl<<"elphi="<<pprim2.elphi<<"=S[s]("<<computeS(msa, s)<<")+"<<N<<"*(Mc+2*Mphi)+"<<"dR("<<delta<<")+AUspsq("<<TP(msa, sp, sq)<<")+AUtxty("<<TP(msa,tx,ty)<<")"<<endl;
			string newlog=ss.str();
			pprim2.log += newlog;
			R.PB(pprim2);
			happen = true;
		} else {
		}
	}	
	return happen;
}


/*******************************************************************************
               COMPUTE ENERGY BASED ON TURNER MODEL
 ******************************************************************************/
int RNAConsT::TP(MSA & in_msa, const int & i1, const int & i2) {
	int ret = 0; 
	FOR(n, 0, in_msa.N) {
		ret += TP(in_msa[n][i1], in_msa[n][i2]);
	}
	return ret; 
}

int RNAConsT::TP(const char & x, const char & y) {
	//terminal penalty;
	if ((x=='U' && (y=='G' || y=='A')) || 
		    ((x=='G' || x=='A') && y=='U')) {
		return TE.MtermianAU;
	}
	return 0;
}

int RNAConsT::d(MSA & in_msa, const int & i1, const int & i2, const int & i3) {
	/*
# 5'-->3'
# d5(h,k-1,k)   d3(i,i+1,j)
# .....\k.....i/-----\
#       |||||||      |
#       h.....j      |
# ...../       \_____/
# d3(h,h+1,k)   d5(i,j-1,j)
# 3'<--5' 
*/
	if(! (i1>=0 && i1<i2 && i2<=in_msa.L-1) ) 
		cout<<"i1="<<i1<<", i2="<<i2<<", i3="<<i3<<endl;
	assert(i1>=0 && i1<i2 && i2<=in_msa.L-1);
	if(i3==-1 || i3==in_msa.L) return 0;
		if(i3==i1-1)	{//k=i1,h=i2;
		int k=i1, h=i2;
		return d5(in_msa, h,k-1,k);
	} else if(i3==i1+1) {
		int i=i1, j=i2;
		return d3(in_msa, i,i+1,j);
	} else if(i3==i2+1) {
		int k=i1, h=i2;
		return d3(in_msa, h,h+1,k);
	} else if(i3==i2-1){
		int i=i1, j=i2;
		return d5(in_msa, i,j-1,j);
	} else {
		assert(false);
	}
}

int RNAConsT::d5(MSA & in_msa, const int & i1, const int & i2, const int & i3) {
	int total = 0;
	FOR(n, 0, in_msa.N) { 	
		string seq = in_msa[n];
		string name = "XYZ";
		name[0] = seq[i1]; name[1] = seq[i2]; name[2] = seq[i3];
		total += TE.d5E[name];
	}
	return total;
}

//seq[i1] pairs with seq[i2]
//#5'-->3'      5'-->3'
//# 1 2             3       ---\ d3(123)
//# 3             2 1       ---/
//#3'<--5'      3'<--5'
int RNAConsT::d3(MSA & in_msa, const int & i1, const int & i2, const int & i3) {
	int total = 0;
	FOR(n, 0, in_msa.N) {
		string seq = in_msa[n];
		string name = "XYZ";
		name[0] = seq[i1]; name[1] = seq[i2]; name[2] = seq[i3];
		total += TE.d3E[name];
	}
	return total;
}

//compute stacking energy of the stem st
int RNAConsT::computeS(MSA & in_msa, stem & st) {
	return computeS(in_msa, st.outer.lp, st.outer.rp, st.inner.lp, st.inner.rp);
}

//compute stacking energy of the stem (x, y, p, q);
int RNAConsT::computeS(MSA & in_msa, const int & x, const int & y, const int & p, const int & q) {
	int total = 0;
	FOR(n, 0, in_msa.N) {
		int e_stacking = 0;
		string seq = in_msa[n];
		FOR(t,0,p-x) {
			int i1 = x+t; int i2 = x+t+1;
			int i3 = y-(t+1); int i4 = y-t;
			//i1,i2,i3,i4 (5'--->3')
			string name = "";
			name += seq[i1]; name += seq[i2];
			name += seq[i3]; name += seq[i4];
			//if the four bases in name can not form a loop, TE.STACKING[name]=0
			e_stacking += TE.STACKING[name];
		}
		total += e_stacking;
	}
	int cov = 0;
	FOR(t, x, p+1) {
		cov += c[t][x+y-t];
	}
	total -= cov; //cov: positive values = strong covariance; negative = bad
	              //energy: positive values = bad values; negative = strong covariance
#ifdef MAINDEBUG
	cout<<"in computeS["<<x<<","<<y<<"~"<<p<<","<<q<<"], Si = "<<total<<endl;	
#endif
	return total;
}

int RNAConsT::computeH(MSA & in_msa, stem & st){
	int N = in_msa.N;
	int L = in_msa.L;

	int x = st.outer.lp; int y = st.outer.rp;
	int p = st.inner.lp; int q = st.inner.rp;
	int Si = computeS(in_msa, st);
#ifdef MAINDEBUG
	cout<<"in computeH["<<x<<","<<y<<"~"<<p<<","<<q<<"], Si = "<<Si;
#endif	
	int total = Si;	
	FOR(n, 0, N) {
		string seq = in_msa[n];
		int hairpinlen = q-p-1;
		int e_loop = 0;
		int energy = 0;
		int e_terminalsmhe = 0; 
		int e_AUpenalty = 0; //DONT count left AU penalty, which should be          \
		calculated when computing H[] or C[][] or FM[][] or FM1[][]
		int e_tetraLoopE = 0;
	
	//internal, bulge or hairpin loops > 30: dS(T)=dS(30)+param*ln(n/30)           \
	--> 1.079, Ref:http://www.bioinfo.rpi.edu/zukerm/cgi-bin/efiles-3.0.cgi#ASINT1x2
		if(hairpinlen>MAXLOOPLEN) {
			e_loop = TE.HAIRPIN[MAXLOOPLEN] + TE.Mlongloop*log(double(hairpinlen)/30);
		} else {
			e_loop = TE.HAIRPIN[hairpinlen];
		}
		if(hairpinlen > 3){
		//if length of the hairpin loop is greater or equal to 4, only count     \
		the terminal mismatch stacking energy, dont count terminal AU GU         \
		penalty (right end)
			string name = "";
			name += seq[p]; name += seq[p+1];
			name += seq[q-1]; name += seq[q];
		 	e_terminalsmhe = TE.TerminalSMHE[name];
		 	//by default, if name is not in TE.TerminalSMHE, then it returns 0 ;

		 	if(hairpinlen==4) {
		 		//check tetra loop energy;
	 			name = seq.substr(p, 6);
	 			map<string, int>::iterator it;
	 			it = TE.TetraLoopE.find(name);
	 			if(it != TE.TetraLoopE.end()) {
	 				e_tetraLoopE = (*it).second;
	 			}
	 			energy = e_terminalsmhe+e_loop+e_tetraLoopE;
				int tmpe = e_tetraLoopE;
				assert(tmpe == e_tetraLoopE);
				//debug here, check if the two ways of computing tetraLoopE is the same	 
			} else {
				energy = e_terminalsmhe+e_loop;
			}
		} else if (hairpinlen==3){
		//if length of the hairpin loop is 3, only count the (right) terminal    \
		AU/GU penalty, do not count terminal mismatch stacking energy.
			e_AUpenalty = TP(seq[p], seq[q]);//TerminalPenalty(TE, seq[p], seq[q]);
			energy = e_AUpenalty+e_loop;
		} else {
		//for single RNA sequence folding, if the loop length is < 3 then, energy = OOE, but for consensus folding, if the loop length of the stem in one of the species is < 3, then we add penalty of 6.00 * UNIT to the overall free energy 
		}
		
		//check for GU closeure preceded by GG
		if(seq[p] == 'G' && seq[q] == 'U') {
			if(p >= 2) {
				if(seq[p-1] == 'G' && seq[p-2] == 'G')
					energy += TE.Mgubous;
			}
		}
		//check for a poly-c loop
		bool polyc_loop = true;
		FOR(k,p+1,q) {
			if(seq[k] != 'C') { 
				polyc_loop = false;
				break;
			}
		}
		if(polyc_loop) {
			if(hairpinlen==3) {
				energy += TE.Mc3;
			} else {
				energy += TE.Mcintercept + TE.Mcslope * hairpinlen;
			}
		}
		total += energy;
	}
	return total;
}

int RNAConsT::computeS(MSA & in_msa, stem & s, stem & t) {
//(((((((((((((.........)))))))))))))))))
//...........((((.....)))))..............
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	int tx = t.outer.lp; int ty = t.outer.rp;
	int tp = t.inner.lp; int tq = t.inner.rp;
	
	assert(s.enclose(t) || t.inward_extend(s));
	if(s.enclose(t)) {
		return computeS(in_msa, sx, sy, sp, sq);
	}
	if(t.inward_extend(s)) {
		assert(ty-sq!=sp-tx);
		if( (sp<tx && tq<=sq && sq<=ty && ty<sy) ||
			 (ty>=sq && sp>=tx && ty-sq > sp-tx)) {
			int star = sx+sy-ty-1;
			return computeS(in_msa, sx, sy, star, ty+1);
		}
		if( (ty<sq && sx<tx && tx<=sp && sp<=tp) ||
			 (ty>=sq && sp>=tx && ty-sq<sp-tx) ){
			int star = sx+sy-tx+1;
			return computeS(in_msa, sx, sy, tx-1, star);
		}
	}
	assert(false);
}

int RNAConsT::computephi(MSA & in_msa, stem & s, stem & t) {
	assert(s.enclose(t) || t.inward_extend(s));

	int e_phi = 0;
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp; 
	int tx = t.outer.lp; int ty = t.outer.rp;
	int tp = t.inner.lp; int tq = t.inner.rp;

	int leftgap  = tx-sp-1; //t.x-s.p-1
	int rightgap = sq-ty-1; //s.q-t.y-1
	int shortgap = (leftgap>rightgap)?(rightgap):(leftgap);
	int longgap  = (leftgap>rightgap)?(leftgap):(rightgap);
	#ifdef MIANDEBUG
	cout<<"longgap="<<longgap<<", shortgap="<<shortgap<<endl;
	#endif
	if(shortgap==0) { //bulge
		e_phi = computeBulge(in_msa, sp, sq, tx, ty);
	} else if(shortgap>0) {
		e_phi = computeInterior(in_msa, sp, sq, tx, ty);
	} else if(shortgap<0) {
		assert(ty-sq!=sp-tx);
		if( (sp<tx && tq<=sq && sq<=ty && ty<sy) ||
			 (ty>=sq && sp>=tx && ty-sq > sp-tx)) {
			int star = sx+sy-ty-1;
			e_phi = computeBulge(in_msa, star, ty+1, tx, ty);
		}
		if( (ty<sq && sx<tx && tx<=sp && sp<=tp) ||
			 (ty>=sq && sp>=tx && ty-sq<sp-tx) ){
			int star = sx+sy-tx+1;
			e_phi = computeBulge(in_msa, tx-1, star, tx, ty);
		}
	}
	return e_phi;
}

int RNAConsT::computeBulge(MSA & in_msa, const int & x, const int & y, 
   const int & p, const int & q) {
//x pairs with y and p pairs with q
//...xp....
//||||\\\\\\
//...y q.....
//   |  \
//   |__/

//   ______
//   |    /
//   |   /
//...x  p.....
//|||| //////
//...yq.....
	int total = 0;
	assert(x<p && p<q && q<y);
	FOR(n, 0, in_msa.N) {
		string seq = in_msa[n];
		int e_bulge = 0;
		int leftgap = p-x-1;
		int rightgap = y-q-1;
		int longgap = (leftgap>rightgap)?(leftgap):(rightgap);
		int shortgap = (leftgap<rightgap)?(leftgap):(rightgap);
		assert(shortgap==0);	
		if(longgap==1) {
			//according to mfold 3.5 algorithm.cpp line 3349
			string name = "";
			name += seq[x]; name += seq[p]; name += seq[q]; name += seq[y];
			e_bulge = TE.BULGE[longgap] + TE.STACKING[name];
		} else if (longgap>30) {
			e_bulge = TE.BULGE[MAXLOOPLEN]+ TE.Mlongloop*log(longgap/MAXLOOPLEN) + 
				   TP(seq[x], seq[y]) + 
				   TP(seq[p], seq[q]);
		} else {
			e_bulge = TE.BULGE[longgap] + 
					TP(seq[x], seq[y]) +
					TP(seq[p], seq[q]);
		}
		total += e_bulge;
	}
	return total;
}

int RNAConsT::computeBulge(MSA & in_msa, stem & s, stem & t) {
	assert(s.enclose(t) || t.inward_extend(s));
	int e_bulge = 0;
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp; 
	int tx = t.outer.lp; int ty = t.outer.rp;
	int tp = t.inner.lp; int tq = t.inner.rp;
	int e_bugle = 0;

	int leftgap  = tx-sp-1; //t.x-s.p-1
	int rightgap = sq-ty-1; //s.q-t.y-1
	int shortgap = (leftgap>rightgap)?(rightgap):(leftgap);
	int longgap  = (leftgap>rightgap)?(leftgap):(rightgap);
	if(shortgap==0) { //bulge
		e_bulge = computeBulge(in_msa, sp, sq, tx, ty);
	} else if(shortgap>0) {
		e_bulge = computeInterior(in_msa, sp, sq, tx, ty);
	} else if(shortgap<0) {
		assert(ty-sq!=sp-tx);
		if( (sp<tx && tq<=sq && sq<=ty && ty<sy) ||
			 (ty>=sq && sp>=tx && ty-sq > sp-tx)) {
			int star = sx+sy-ty-1;
			e_bulge = computeBulge(in_msa, star, ty+1, tx, ty);
		}
		if( (ty<sq && sx<tx && tx<=sp && sp<=tp) ||
			 (ty>=sq && sp>=tx && ty-sq<sp-tx) ){
			int star = sx+sy-tx+1;
			e_bulge = computeBulge(in_msa, tx-1, star, tx, ty);
		}
	}
	return e_bulge;
}

int RNAConsT::computeInterior(MSA & in_msa, const int & x, const int & y, const int & p, const int & q) {
	int total = 0;
	FOR(n, 0, in_msa.N) {
		int e_interior = 0;
		string seq = in_msa[n];
		assert(x<p && p<q && q<y);
		int leftgap = p-x-1;
		int rightgap = y-q-1;
		int longgap = (leftgap>rightgap)?(leftgap):(rightgap);
		int shortgap = (leftgap<rightgap)?(leftgap):(rightgap);
		assert(shortgap!=0 && longgap!=0);
		int allgap = shortgap+longgap;
		if(leftgap==1 && rightgap==1) {
			string name = "";
			name += seq[x]; name += seq[x+1]; name += seq[x+2];
			name += seq[y-2]; name += seq[y-1]; name += seq[y];
			e_interior = TE.SingleMIE[name];
		} else if(leftgap==1 && rightgap==2) {
			string name = "";
			name += seq[x]; name += seq[x+1]; name += seq[x+2];
			name += seq[y-3]; name += seq[y-2]; name += seq[y-1]; name += seq[y];
			e_interior = TE.OneTwoMIE[name];
		} else if(leftgap==2 && rightgap==1) {
			string name = "";
			name += seq[y-2]; name += seq[y-1]; name += seq[y];
			name += seq[x]; name += seq[x+1]; name += seq[x+2]; name += seq[x+3];
			e_interior = TE.OneTwoMIE[name];
		} else if (leftgap==2 && rightgap==2) {
			string name = "";
			name = seq.substr(x, 4);
			name += seq[y-3]; name += seq[y-2]; name += seq[y-1]; name += seq[y];
			e_interior = TE.TwoTwoMIE[name];
		}

		if(allgap > MAXLOOPLEN) {
			e_interior = TE.INTERNAL[MAXLOOPLEN]+ TE.Mlongloop*log(allgap/MAXLOOPLEN);
			e_interior += (TE.Mninio_max < (TE.Mninio_m*(longgap-shortgap)))? (TE.Mninio_max): (TE.Mninio_m*(longgap-shortgap));
		} else if(allgap > 4 || (allgap==4 && longgap==3 && shortgap==1)) {
			e_interior = TE.INTERNAL[allgap];
			e_interior += min(TE.Mninio_max, (TE.Mninio_m*(longgap-shortgap)));
		}
		if(shortgap==1 && longgap > 2) {
		//according to mfold, assume we had a loop composed of AA between
			string name = "";
			name += seq[x]; name += 'A'; name += 'A'; name += seq[y];
			e_interior  += TE.TerminalSMIE[name];
			name = ""; 
			name += seq[q]; name += 'A'; name += 'A'; name += seq[p];
			e_interior  += TE.TerminalSMIE[name];
		}
		if(shortgap>=2 && longgap > 2) {
			string name = "";
			name += seq[x]; name += seq[x+1]; name += seq[y-1]; name += seq[y];
			e_interior  += TE.TerminalSMIE[name];
			name = "";
			name += seq[q]; name += seq[q+1]; name += seq[p-1]; name += seq[p];
			e_interior  += TE.TerminalSMIE[name];
		}
		total += e_interior;
	}
	return total;
}


stem RNAConsT::getUsedPartOfStem(stem & s, stem & t) {
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	int tx = t.outer.lp; int ty = t.outer.rp;
	int tp = t.inner.lp; int tq = t.inner.rp;
	assert(s.enclose(t) || t.inward_extend(s));
	if(s.enclose(t)) {
		return s;
	}
	if(t.inward_extend(s)) {
		assert(ty-sq!=sp-tx);
		if( (sp<tx && tq<=sq && sq<=ty && ty<sy) ||
			 (ty>=sq && sp>=tx && ty-sq > sp-tx)) {
			int star = sx+sy-ty-1;
			// computeBulge(seq, star, ty+1, tx, ty);
			// computeS(seq, sx, sy, star, ty+1);
			stem ret(BP(sx, sy), BP(star, ty+1));
			assert(ret.outer.lp <= ret.inner.lp && ret.inner.lp < ret.inner.rp && ret.inner.rp <= ret.outer.rp);
			return ret;
		}
		if( (ty<sq && sx<tx && tx<=sp && sp<=tp) ||
			 (ty>=sq && sp>=tx && ty-sq<sp-tx) ) {
			int star = sx+sy-tx+1;
			//computeBulge(seq, tx-1, star, tx, ty);
			//computeS(seq, sx, sy, tx-1, star);
			stem ret(BP(sx, sy), BP(tx-1, star));
			assert(ret.outer.lp <= ret.inner.lp && ret.inner.lp < ret.inner.rp && ret.inner.rp <= ret.outer.rp);
			return ret;
		}
	}
	assert(false);

}

stem RNAConsT::getUsedPartOfStem(stemlist & in_stl, const int & indx_s, const int & indx_t) {
 //s.enclose(t) or t.inward_extend(s), return the used part of S
	stem s = in_stl.vstems[indx_s];
	stem t = in_stl.vstems[indx_t];
	return getUsedPartOfStem(s, t);
}

//...(((((......)))))........... s
//.................((((.....)))) t
stem RNAConsT::getUsedPartOfStem_Parallel(stem & s, stem & t) {
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	int tx = t.outer.lp; int ty = t.outer.rp;
	int tp = t.inner.lp; int tq = t.inner.rp;
	
	assert(sy<tx || (sq<tx && tx<=sy && sy<=tp));
	return stem(BP(sx, sy), BP(sx+sy-tx+1, tx-1));
}


