#ifndef _RNACONS_TURNER_H_
#define _RNACONS_TURNER_H_

#include "Util.h"
#include "Turner.h"
#include "TurnerParam.h"
#include "RNAStruct.h"
#include "MSA.h"
#include <map>

class stem; 
class stemlist;
class PartStackStructT;
class RNA;


/*************************CLASS SCORE*********************************/
class SCORE {
public: 
	int ec; //energy score + covariance score
	int e;  //energy score;
	SCORE() {ec = e = 0;}
	SCORE(const int & energy_covariance , const int & energy) {
		ec = energy_covariance; e = energy; 
	}
	double getcovariance(){return ec-e;}
	void show(ostream & fout=cout) {fout<<"("<<ec<<", "<<e<<")"<<endl;}
};
SCORE operator + (const SCORE & sc1, const SCORE & sc2);
SCORE operator - (const SCORE & sc1, const SCORE & sc2);
//Assume < and > is determined by the first variable ec
bool operator < (const SCORE & sc1, const SCORE & sc2);
bool operator > (const SCORE & sc1, const SCORE & sc2);
bool operator == (const SCORE & sc1, const SCORE & sc2); 

/*************************ENUM CTPYE*****************************/
//Case types
enum CTYPE {
	ctype_unknown = -1,
	ctype_0 = 0,
	ctype_1 = 1,
	ctype_2 = 2,
	ctype_3 = 3,
	ctype_4 = 4,
	ctype_1a = 5,
	ctype_1b = 6,
	ctype_1c = 7,
	ctype_2a = 8,
	ctype_2b = 9,
	ctype_2c = 10,	
	ctype_3a = 11, //M4[i1][j1]
	ctype_3b = 12, //M4[i1][j2]
	ctype_3c = 13, //M4[i1][j3]
	ctype_3d = 14, //M4[i2][j1]
	ctype_3e = 15, //M4[i2][j2]
	ctype_3f = 16, //M4[i2][j3]
	ctype_3g = 17, //M4[i3][j1]
	ctype_3h = 18, //M4[i3][j2]
	ctype_3i = 19, //FM[i3][j3]
	ctype_3p = 20, //M2[i3][j1]
	ctype_3q = 21, //M2[i3][j2]
	ctype_3r = 22, //M1[i1][j3]
	ctype_3s = 23, //M1[i2][j3]
	ctype_4a = 24,
	ctype_4b = 25,
	ctype_5a = 26,
	ctype_5b = 27
};
void showCTYPE(const CTYPE & ctype, ostream & fout=cout) ;
/*************************CLASS StackFoldAuxiliary*****************************/
class StackFoldAuxiliary{
	public:
	CTYPE ctype;
	stem st;
	
	StackFoldAuxiliary();
	StackFoldAuxiliary(const CTYPE & in_ctype, const stem & in_t);
	StackFoldAuxiliary(const StackFoldAuxiliary & in_sfa);
//	~StackFoldAuxiliary();
	void set(const CTYPE & in_ctype, const stem & t);
	void show(ostream & fout = cout);
};

/*************************CLASS RNAConsT*********************************/
class RNAConsT{
public:
	MSA msa;
	stemlist stl;//predicted set of stems in a stemlist;
	int N; //number of sequences
	int L; //length of sequences
	
	map<char, int> base; //Convert a base to an interger, _=0, A=1, C=2, G=3, U=4;
	int pair[5][5];//Code base pair	
	
	vector<vector<int>> c; //covariance matrix * UNIT
	vector<vector<bool>> ali_pi;//in the consensus structure, base i and j are pairing or not
	
//	vector<int> H; //map the index of the stem in the stemlist to (energy*UNIT)
//	map<int, int> S; //map the index of two stems in the stemlist to (energy*UNIT)
//	map<int, int> phi; //map the indice of two stems in the stemlist to (energy*UNIT)
	
	vector<int> C;  //C[i][j] = energy of a segment [i,j] provided that the segment is enclosed by a stem with outer=(i,j)
	vector<int> FM; //FM[i][j] = energy of segment [i,j], which is a part of a multi-loop and it contains at least one stem
	vector<int> M1; //M1[i][j] = energy of segment [i,j], which is a part of a multi-loop and it contains at least one stem starting at i
	vector<int> M2; //M2[i][j] = energy of segment [i,j], which is a part of a multi-loop and it contains at least one stem ending at j
	vector<int> M4; //M4[i][j] = energy of segment [i,j], which is a part of a multi-loop and it contains at least one stem starting at i and one stem ending at j (the stem starting at i and the stem ending at j can be identical)
	
	vector<int> F;
	vector<int> Fn;
	vector<int> Fd;
	vector<int> F2n;
	vector<int> F2d;

	vector<StackFoldAuxiliary> F_sfa; //Auxiliary
	vector<StackFoldAuxiliary> Fn_sfa; //Auxiliary  
	vector<StackFoldAuxiliary> Fd_sfa; //Auxiliary  
	vector<StackFoldAuxiliary> F2n_sfa; //Auxiliary  
	vector<StackFoldAuxiliary> F2d_sfa; //Auxiliary  
	
	vector<StackFoldAuxiliary> C_sfa;
	vector<StackFoldAuxiliary> FM_sfa;
	vector<StackFoldAuxiliary> M1_sfa;
	vector<StackFoldAuxiliary> M2_sfa;
	vector<StackFoldAuxiliary> M4_sfa;

	vector<stem> F_S; //the right most stem in F_S
	vector<stem> Fn_S;
	vector<stem> Fd_S;
	vector<stem> F2n_S;
	vector<stem> F2d_S;

	vector<stem> FM_LS; //the left most stem in FM[i,j]
	vector<stem> FM_RS; //the right most stem in FM[i,j]
	vector<stem> M1_LS; //the left most stem in M1[i,j]
	vector<stem> M1_RS; //the right most stem in M1[i,j]	
	vector<stem> M2_LS; //the left most stem in M2[i,j]
	vector<stem> M2_RS; //the righStackFoldt most stem in M2[i,j]	
	vector<stem> M4_LS; //the left most stem in M4[i,j]
	vector<stem> M4_RS; //the right most stem in M4[i,j]

	RNAConsT(){}
	RNAConsT(const RNAConsT & in_rc);
	RNAConsT(const MSA & in_msa);
	void init();
	void compute_covariance();
	void init_stl();
	//Stack based consensus folding and subopt structure generation
//	SCORE calscore(PartStackStruct & phi);
//	SCORE stackfold(stemlist & in_stl);
//	string stacktraceback(stemlist & in_stl, const bool & is_countec);
//	vector<RNA> stacksubopt(stemlist & in_stl, ostream & fout = cout, const double & delta = 1, const bool & is_coutec = true);
	void show_c();	
	void show_ali_pi();
	int stackfold();
	string traceback();
	vector<RNA> generate_subopt(ofstream & fout, int & e_ub, bool & reqLocOpt);
	int computeEnergy(stemlist & in_stl);
	int computeScov(stemlist & in_stl);	
	
	/* functions for computing energy for multiple alignment: begin */
	int TP(MSA & in_msa, const int & i1, const int & i2);
	int TP(const char & x, const char & y);//terminal penalty;
	int d(MSA & in_msa, const int & i1, const int & i2, const int & i3);
	int d5(MSA & in_msa, const int & i1, const int & i2, const int & i3);
	int d3(MSA & in_msa, const int & i1, const int & i2, const int & i3);
	//int computeEnergy(const string & seq, RStack & in_stl); //given a secondary structure, predict its energy.
	int computeS(MSA & in_msa, stem & st);
	int computeH(MSA & in_msa, stem & st);
	int computeBulge(MSA & in_msa, const int & x, const int & y, const int & p, const int & q);
	int computeBulge(MSA & in_msa, stem & s, stem & t);
	int computeInterior(MSA & in_msa, const int & x, const int & y, const int & p, const int & q);
	int computeS(MSA & in_msa, stem & s, stem & t);
	int computeS(MSA & in_msa, const int & x, const int & y, const int & p, const int & q);
	int computephi(MSA & in_msa, stem & s, stem & t);
	stem getUsedPartOfStem(stem & s, stem & t);
	stem getUsedPartOfStem(stemlist & in_stl, const int & indx_s, const int & indx_t); //s.enclose(t) or t.inward_extend(s), return the used part of S
	stem getUsedPartOfStem_Parallel(stem & s, stem & t);
	/* functions for computing energy for multiple alignment: end */
private:
	void fillH();
	void fillS();
	void fillphi();
	
	void fillC(const int & i, const int & j);
	void fillFM(const int & i, const int & j);
	void fillM1(const int & i, const int & j);
	void fillM2(const int & i, const int & j);
	void fillM4(const int & i, const int & j);
	
	void fillF2n2d(const int & i);
	void fillFnFd(const int & i);
	void fillF(const int & i);
	
	int pos(const int & i, const int & j);
	int getE(const ETYPE & etype, const int & i, const int & j=-1);
	stem getStem(const vector<stem> & vstem, const int & i);
	stem getStem(const vector<stem> & vstem, const int & i, const int & j);
	

	vector<stem> tracebackF(const int & j);
	vector<stem> tracebackFn(const int & j);
	vector<stem> tracebackFd(const int & j);
	vector<stem> tracebackF2n(const int & j);
	vector<stem> tracebackF2d(const int & j);
	vector<stem> tracebackF2nF2d(const int & j, StackFoldAuxiliary & sfa);
	vector<stem> tracebackC(const int & i, const int & j);
	vector<stem> tracebackFM(const int & i, const int & j);
	vector<stem> tracebackM1(const int & i, const int & j);
	vector<stem> tracebackM2(const int & i, const int & j);
	vector<stem> tracebackM4(const int & i, const int & j);
	
	int getsegment_esum(seglistT & in_sgl);

	bool suboptF(vector<PartStackStruct> & R, const segmentT & in_sg, const PartStackStruct & in_phi, const int & e_ub, const bool & reqLocOpt);
	bool suboptFM(vector<PartStackStruct> & R, const segmentT & in_sg, const PartStackStruct & in_phi, const int & e_ub, const bool & reqLocOpt);
	bool suboptC(vector<PartStackStruct> & R, const segmentT & in_sg, const PartStackStruct & in_phi, const int & e_ub, const bool & reqLocOpt);
	

};




/*
int TerminalPenalty(TurnerE & TE, const char & x, const char & y);
int d(TurnerE & TE, const string & seq, const int & i1, const int & i2, const int & i3);
int d5(TurnerE & TE, const string & seq, const int & i1, const int & i2, const int & i3);
int d3(TurnerE & TE, const string & seq, const int & i1, const int & i2, const int & i3);
int computeEnergy(TurnerE & TE, const string & seq, RStack & in_stl); //given a secondary structure, predict its energy.
int computeH(TurnerE & TE, const string & seq, stem & st);
int computeBulge(TurnerE & TE, const string & seq, const int & x, const int & y, const int & p, const int & q);
int computeInterior(TurnerE & TE, const string & seq, const int & x, const int & y, const int & p, const int & q);
int computeS(TurnerE & TE, const string & seq, const int & x, const int & y, const int & p, const int & q);
int computeS(TurnerE & TE, const string & seq, stem & s, stem & t);
int computeS(TurnerE & TE, const string & seq, stem & st);
int computephi(TurnerE & TE, const string & seq, stem & s, stem & t);
stem getUsedPartOfStem(stem & s, stem & t);
stem getUsedPartOfStem(stemlist & in_stl, const int & indx_s, const int & indx_t); //s.enclose(t) or t.inward_extend(s), return the used part of S
*/



#endif
