#include "RNAStruct.h"
#include "Config.h"
extern Config cf; 

/*******************************************************************************
                        enum ETYPE
 ******************************************************************************/
void showETYPE(const ETYPE & etype, ostream & fout) {
/*Types of Energy contain both the Nussinov model and the Turner model 
	Nussinove Model: P0, P1
	Turner Model: F, C, FM
	0 - unknown energy type
*/
	switch(etype) {
		case etype_unknown: {fout<<"unknown";break;}
		case etype_F:{fout<<"F ";break;}
		case etype_C:{fout<<"C ";break;}
		case etype_FM:{fout<<"FM ";break;}
		case etype_P0: {fout<<"P0 "; break;}
		case etype_P1: {fout<<"P1 "; break;}
		default:{cerr<<"unknown ETYPE="<<etype<<endl;exit(0);}
	}
}

/***********************************************************
                    overload operator for BP
***********************************************************/
bool operator == (const BP & in_bp1, const BP & in_bp2) {
	return ( (in_bp1.lp==in_bp2.lp && in_bp1.rp==in_bp2.rp) 
//	||       (in_bp1.lp==in_bp2.rp && in_bp1.rp==in_bp2.lp) 
	);
}
bool operator != (const BP & in_bp1, const BP & in_bp2) {
	return ( in_bp1.lp != in_bp2.lp ||  in_bp1.rp != in_bp2.rp);
}

bool operator < (const BP & in_bp1, const BP & in_bp2) {
	return ( in_bp1.lp < in_bp2.lp ||
	         ((in_bp1.lp==in_bp2.lp) && (in_bp1.rp > in_bp2.rp) ));
}

bool operator > (const BP & in_bp1, const BP & in_bp2) {
	return ( in_bp1.lp > in_bp2.lp ||
	         ((in_bp1.lp==in_bp2.lp) && (in_bp1.rp < in_bp2.rp) ));
}


/***********************************************************
                  overload operator for stem
***********************************************************/
bool operator == (const stem & st1, const stem & st2) {
	return (st1.outer.lp == st2.outer.lp && st1.outer.rp == st2.outer.rp
	     && st1.inner.lp == st2.inner.lp && st1.inner.rp == st2.inner.rp);
}
bool operator != (const stem & st1, const stem & st2) {
	return (st1.outer.lp != st2.outer.lp || st1.outer.rp != st2.outer.rp 
	     || st1.inner.lp != st2.inner.lp || st1.inner.rp != st2.inner.rp);
}

bool operator < (const stem & st1, const stem & st2) {
	return (st1.outer.lp < st2.outer.lp || 
	       (st1.outer.lp == st2.outer.lp && st1.outer.rp > st2.outer.rp));
}
bool operator > (const stem & st1, const stem & st2) {
	return (st1.outer.lp > st2.outer.lp ||
          (st1.outer.lp == st2.outer.lp && st1.outer.rp < st2.outer.rp));
}


/***********************************************************
                    overload operators for RNA
***********************************************************/
bool operator < (RNA & rna1, RNA & rna2) {
	return (rna1.e < rna2.e);
}

bool operator > (RNA & rna1, RNA & rna2) {
	return (rna1.e > rna2.e);

}

bool descendingsortRNAbyE(RNA rna1, RNA rna2) {	
	return (rna1.e > rna2.e);
}
bool ascendingsortRNAbyE(RNA rna1, RNA rna2) {
	return (rna1.e < rna2.e);
}


bool descendingsortRNAbyC(RNA rna1, RNA rna2) {	
	return (rna1.c > rna2.c);
}
bool ascendingsortRNAbyC(RNA rna1, RNA rna2) {
	return (rna1.c < rna2.c);
}

/***********************************************************
                         stem
***********************************************************/
bool stem::inward_extend(const stem & s) {
//(((((((((............)))))))))) --> stem s
//.....((((((.......))))))....... --> this stem
//or 
//(((((((((............)))))))))) --> stem s
//.........((((((....))))))...... --> this stem
//or
//(((((((((............)))))))))) --> stem s
//....((((((....))))))........... --> this stem
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	int tx = outer.lp; int ty = outer.rp;
	int tp = inner.lp; int tq = inner.rp;
	bool ret = false;
	if( (sx<tx && tx<=sp && sp<=tp && tq<=sq && ty<sy && !enclosedby(s)) ||
	    (sx<tx && sp<=tp && tq<=sq && sq<=ty && ty<sy && !enclosedby(s)) )
	{
		int leftoverlap = sp-tx+1;
		int rightoverlap = ty-sq+1;
		if((leftoverlap <= cf.MAXOVERLAP ) && 
		   (rightoverlap <= cf.MAXOVERLAP) ) {
		   return true;
		} else {
			return false;
		}
	} else return false;
}

bool stem::prevparallel_extend(const stem & s) {
//....................((((((............))))))  stem s
//((((((.........)))))))) this stem
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	int tx = outer.lp; int ty = outer.rp;
	int tp = inner.lp; int tq = inner.rp;
//	if(tq<sx && sx<=ty && ty<sp && ty-sx+1<=cf.MAXPARALLELOVERLAP) {
	if(tq<sx && sx<=ty && ty<sp && ty-sx+1<=cf.MAXOVERLAP) {
		return true;
	} else return false;
}

bool stem::parallel_extend(const stem & s) {
//((((((.........)))))))) stem s
//....................(((............))))))  this stem

//..............((((((((.......)))))))) stem s 
//..(((((.....))))))).................. this stem
	
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	int tx = outer.lp; int ty = outer.rp;
	int tp = inner.lp; int tq = inner.rp;
	
//	if( (sq<tx && tx<=sy && sy<tp && (sy-tx+1) <= cf.MAXPARALLELOVERLAP) || 
//	    (tq<sx && sx<=ty && ty<sp && (ty-sx+1) <= cf.MAXPARALLELOVERLAP) ) {
	if( (sq<tx && tx<=sy && sy<tp && (sy-tx+1) <= cf.MAXOVERLAP) || 
	    (tq<sx && sx<=ty && ty<sp && (ty-sx+1) <= cf.MAXOVERLAP) ) {
		return true;
	} else return false;
}

int stem::inwhichseg(const int & query) {
	//stem st (x,p) ~ (q,y) is composed of five segments, including 
	//segment 1 : [0..x-1]
	//segment 2 : [x..p]
	//segment 3 : [p+1,q-1]
	//segment 4 : [q..y]
	//segment 5 : [y+1..+OO]
	//return in which segment is query
	int x = outer.lp; int y = outer.rp;
	int p = inner.lp; int q = inner.rp;
	
	if(query<x) {
		return 1; 
	} else if(x<=query && query<=p) {
		return 2;
	} else if(p<query  && query<q ) {
		return 3;
	} else if(q<=query && query<=y) {
		return 4;
	} else if(query>y) {
		return 5;
	} else {
		assert(false);
	}
}

/***********************************************************
                         stemlist
***********************************************************/
stemlist::stemlist(const string & seq) {
	L = sz(seq);
	vector<vector<bool>> pi;
	FOR(i,0,L) {
		vector<bool> vb; 
		FOR(j,0,L) {
			vb[j] = false;
		}
		pi.PB(vb);
	}
	FOR(i,0,L) 
		FOR(j,0,L) {
			if(match(seq[i], seq[j]) && j-i>cf.MINLOOPLEN) 
				pi[i][j] = true;
		}
	construct_usingpi(pi);
	removestem_byhydrobonds(seq);
}

stemlist::stemlist(vector<vector<bool>> & pi, MSA & in_msa) {
	L = sz(pi);
	construct_usingpi(pi);
	removestem_byhydrobonds(in_msa);
}

void stemlist::removestem_byhydrobonds(const string & seq) {
	FOR(i,0,sz(vstems)) {
		if(vstems[i].gethydrobonds(seq) < cf.MINHYDROGEN) {
			vstems.erase(vstems.begin()+i);
			i--;
		}
	}
}

void stemlist::removestem_byhydrobonds(MSA & in_msa) {
	FOR(i,0,sz(vstems)) {
		if(vstems[i].gethydrobonds(in_msa) < in_msa.N*cf.MINHYDROGEN) {
			vstems.erase(vstems.begin()+i);
			i--;
		}
	}
}


void stemlist::construct_usingpi(vector<vector<bool>> & pi) {
	int n = sz(pi);
	if(n==0) return ;
	assert(n==sz(pi[0]));
	FOR(k, 0, n) FOR(t, 0, 2) {
		int i, j, start_i, start_j; 
		i = k; j = k+t;
		start_i = start_j = -1;
		while(true) {
			if(i<0 || j>n-1) break;
			if(j-i<=cf.MINLOOPLEN) {
				pi[i][j] = false;
				i--, j++; continue;
			}
			if(!pi[i][j]) {
				if(start_i!=-1 && start_j!=-1) {
					int stacklen = start_i-(i+1)+1; 
					stem newst(BP(i+1,j-1), BP(start_i, start_j)); 
					if(stacklen >= cf.MINSTACKLEN) {
						vstems.PB(newst);
					} else {//
						FOR(x, i+1, start_i+1) {
							pi[x][i+j-x] = false;
						}
					}
				}
				start_i = start_j = -1;
			} else {//pi[i][j]=true
				if(i==0 || j==n-1) {
					if(start_i!=-1 && start_j!=-1) {
						int stacklen = start_i-i+1;
						stem newst(BP(i,j), BP(start_i, start_j));
						if(stacklen >= cf.MINSTACKLEN) {
							vstems.PB(newst);
						} else {
							FOR(x, i, start_i+1) {
								pi[x][i+j-x] = false;
							}
						}
					} else {	//start_i=start_j=-1;
						pi[i][j] = false;
					}
				} else {
					if(start_i==-1 && start_j==-1) {
						start_i = i; start_j = j;
					}
				}
			}
			i--, j++; 
		}
	}
	sort(vstems.begin(), vstems.end(), ascendingsortST);
}

//Note that pi is different from itable
stemlist::stemlist(const int & L, const double & cutoff, const vector<vector<double>> & b, vector<vector<int>> & pi) {
	FOR(k, 0, L) FOR(t, 0, 2) {
		int i, j, start_i, start_j; 
		i = k; j = k+t;
		start_i = start_j = -1;
		while(true) {
			if(i<0 || j>L-1) break;
			if(j-i<=3) {
				pi[i][j] = 0;
				i--, j++; continue;
			}
			if(pi[i][j]==0) {
				if(start_i!=-1 && start_j!=-1) {
					int stacklen = start_i-(i+1)+1; 
					if(stacklen >= cf.MINSTACKLEN) {
						stem newst(BP(i+1,j-1), BP(start_i, start_j)); 
						vstems.PB(newst);
					} else {//
						FOR(x, i+1, start_i+1) {
							pi[x][i+j-x] = 0;
						}
					}
				}
				start_i = start_j = -1;
			} else {//pi[i][j]=1
				if(i==0 || j==L-1) {
					if(start_i!=-1 && start_j!=-1) {
						int stacklen = start_i-i+1;
						if(stacklen >= cf.MINSTACKLEN) {
							stem newst(BP(i,j), BP(start_i, start_j));
							vstems.PB(newst);
							cout<<"PB2 "; newst.show(); cout<<endl;
						} else {
							FOR(x, i, start_i+1) {
								pi[x][i+j-x] = 0;
							}
						}
					} else {	//start_i=start_j=-1;
						pi[i][j] = 0;
					}
				} else {
					if(start_i==-1 && start_j==-1) {
						start_i = i; start_j = j;
					}
				}
			}
			i--, j++; 
		}
	}
	sort(vstems.begin(), vstems.end(), ascendingsortST);
}
void stemlist::show() {
	FOR(i,0,sz(vstems)) {
		cout<<"stem ["<<i<<"/"<<sz(vstems)-1<<"] :";
		vstems[i].show();
		cout<<endl;
	}
	vector<vector<int>> vvi; 

	FOR(i,0,L) {
		vector<int> vi; 
		FOR(j,0,L){
			vi.PB(0);
		}
		vvi.PB(vi);
	}
	
	FOR(i,0,sz(vstems)) {
		int x = vstems[i].outer.lp; int y = vstems[i].outer.rp; 
		int p = vstems[i].inner.lp; int q = vstems[i].inner.rp; 
		FOR(k,x,p+1) {
			vvi[k][x+y-k] = 1;
		}
	}
	string comment = ""; 
	comment.insert(0, L, '+');
	
	cout<<comment<<endl;
	FOR(i,0,L) {
		FOR(j,0,L) {
			if(vvi[i][j]==0) cout<<" ";
			else cout<<".";
		}
		cout<<endl;
	}
	cout<<endl;
	cout<<comment<<endl;
}

int stemlist::getstem_index(const stem & in_st) {
	//return index of in_st in vstems
	int low = 0; int high = sz(vstems);
	while(low <= high) {
		int mid = (low+high)/2;
		if(in_st == vstems[mid]) {
			return mid;
		} else if(in_st < vstems[mid]) {
			high = mid-1;
		} else {
			low = mid+1;
		}
	}
	return -1;
}

//return stacks starting from s
vector<stem> stemlist::getstems_startfrom(int in_s) {
	vector<stem> ret;
	int i=0; int j=sz(vstems)-1;
	int find = -1;
	while(i<=j) {
		int mid = (i+j)/2;
		if(vstems[mid].outer.lp == in_s) {
			find = mid;
			break;
		} else if (vstems[mid].outer.lp > in_s) {
			j = mid-1;
		} else {
			i = mid+1;
		}
	}
	if(find == -1) return ret;
//	ret.PB(vstems[find]);
	int cursor = find-1;
	int start = find;
	int end = find;
	while(cursor>=0) {
		if(vstems[cursor].outer.lp == in_s) start=cursor;//ret.PB(vstems[cursor]);
		else break;
		cursor--;
	}
	
	cursor = find+1;
	while(cursor<sz(vstems)) {
		if(vstems[cursor].outer.lp ==in_s) end=cursor;
		else break;
		cursor++;
	}
	FOR(i,start, end+1) ret.PB(vstems[i]);
	
	FOR(i,0,sz(ret)) {
		assert(ret[i].outer.lp == in_s);
	}
	return ret;
}


//return stacks starting from s
vector<int> stemlist::getstemindice_startfrom(int in_s) {
	vector<int> ret;
	int i=0; int j=sz(vstems)-1;
	int find = -1;
	while(i<=j) {
		int mid = (i+j)/2;
		if(vstems[mid].outer.lp == in_s) {
			find = mid;
			break;
		} else if (vstems[mid].outer.lp > in_s) {
			j = mid-1;
		} else {
			i = mid+1;
		}
	}
	if(find == -1) return ret;
//	ret.PB(vstems[find]);
	int cursor = find-1;
	int start = find;
	int end = find;
	while(cursor>=0) {
		if(vstems[cursor].outer.lp == in_s) start=cursor;//ret.PB(vstems[cursor]);
		else break;
		cursor--;
	}
	
	cursor = find+1;
	while(cursor<sz(vstems)) {
		if(vstems[cursor].outer.lp == in_s) end=cursor;
		else break;
		cursor++;
	}
	FOR(i,start, end+1) ret.PB(i);
	
	FOR(i,0,sz(ret)) {
		assert(vstems[ret[i]].outer.lp == in_s);
	}
	return ret;
}

//return stacks ending at e
vector<stem> stemlist::getstems_endat(int in_e) {
	vector<stem> ret;
	if(in_e<0) return ret;
	FOR(i,0,sz(vstems)) {
		if(vstems[i].outer.rp == in_e) {
			ret.PB(vstems[i]);
		}
	}
	FOR(i,0,sz(ret)) {
		assert(ret[i].outer.rp==in_e);
	}
	return ret;
}

vector<int> stemlist::getstemindice_endat(int in_e) {
	vector<int> ret;
	if(in_e<0) return ret;
	FOR(i,0,sz(vstems)) {
		if(vstems[i].outer.rp == in_e) {
			ret.PB(i);
		}
	}
	FOR(i,0,sz(ret)) {
		assert(vstems[ret[i]].outer.rp==in_e);
	}
	return ret;
}

vector<stem> stemlist::getstems_startfrom_endat(int in_s, int in_e){
	vector<stem> ret;
	if(in_s<0 || in_e<0) return ret;
	vector<stem> vst = getstems_startfrom(in_s);
	FOR(i,0,sz(vst)){
		if(vst[i].outer.rp == in_e) {
			ret.PB(vst[i]);
		}
	}
	FOR(i,0,sz(ret)) {
		assert(ret[i].outer.lp==in_s && ret[i].outer.rp==in_e);
	}
	return ret;
}

	
vector<int> stemlist::getstemindice_startfrom_endat(int in_s, int in_e){
	vector<int> ret;
	if(in_s<0 || in_e<0) return ret;
	vector<int> vi = getstemindice_startfrom(in_s);
	FOR(i,0,sz(vi)){
		if(vstems[vi[i]].outer.rp == in_e) {
			ret.PB(vi[i]);
		}
	}
	FOR(i,0,sz(ret)) {
		assert(vstems[ret[i]].outer.lp == in_s && vstems[ret[i]].outer.rp == in_e);
	}
	return ret;
}

vector<int> getsubstemindice_startfrom_endat(int in_s, int in_e) {
}
	

vector<stem> stemlist::getstems_within(int in_s, int in_e){
	vector<stem> ret;
	FOR(i,0,sz(vstems)){
		if(vstems[i].outer.lp >= in_s && vstems[i].outer.rp <= in_e){
			ret.PB(vstems[i]);
		}
	}
	FOR(i,0,sz(ret)) {
		assert(ret[i].outer.lp >= in_s && ret[i].outer.rp <= in_e);
	}
	return ret;
}

vector<int> stemlist::getstemindice_within(int in_s, int in_e){
	vector<int> ret;
	FOR(i,0,sz(vstems)){
		if(vstems[i].outer.lp >= in_s && vstems[i].outer.rp <= in_e){
			ret.PB(i);
		}
	}
	FOR(i,0,sz(ret)) {
		assert(vstems[ret[i]].outer.lp >= in_s && vstems[ret[i]].outer.rp <= in_e);
	}
	return ret;
}

vector<stem> stemlist::getstems_startfrom_within(int in_s, int in_e) {
	vector<stem> ret;
	vector<stem> vst;
	vst = getstems_startfrom(in_s); 
	FOR(i,0,sz(vst)) {
		if(vst[i].outer.rp <= in_e) ret.PB(vst[i]);
	}
	FOR(i,0,sz(ret)) {
		assert(ret[i].outer.lp == in_s && ret[i].outer.rp <= in_e);
	}
	return ret;
}

vector<int> stemlist::getstemindice_startfrom_within(int in_s, int in_e) {
	vector<int> ret;
	vector<int> vi = getstemindice_startfrom(in_s);
	FOR(i,0,sz(vi)) {
		if(vstems[vi[i]].outer.rp <= in_e) ret.PB(vi[i]);
	}
	FOR(i,0,sz(ret)) {
		assert(vstems[ret[i]].outer.lp == in_s && vstems[ret[i]].outer.rp <= in_e);
	}
	return ret;
}

vector<int> stemlist::getstemindice_endat_within(int in_s, int in_e) {
	vector<int> ret;
	vector<int> vi = getstemindice_endat(in_e);
	FOR(i,0,sz(vi)) {
		if(vstems[vi[i]].outer.lp >= in_s) ret.PB(vi[i]);
	}
	FOR(i,0,sz(ret)) {
		assert(vstems[ret[i]].outer.lp >= in_s && vstems[ret[i]].outer.rp == in_e);
	}
	return ret;
}

vector<int> stemlist::getstemindice_enclosedby(int indx_s) {
	vector<int> ret;
	stem s = vstems[indx_s];
	FOR(i,indx_s+1,sz(vstems)){
		stem sti = vstems[i];
		if(sti.outer.lp >= s.outer.rp) break;
		if(s.enclose(sti)) ret.PB(i);
	}
	return ret;
}

//	(((((((.......)))))))  -->stem s
//    ((((((....))))))    -->stem t

// (((((((.......)))))))  -->stem s
//    ((((.....))))       -->stem t

// (((((((.......)))))))  -->stem s
//       ((((...))))      -->stem t

// (((((((.........)))))))-->stem s
//       ((((....))))     -->stem t
vector<int> stemlist::getstemindice_inward_extend(int indx_s) {
	vector<int> ret;
	stem s = vstems[indx_s];
	FOR(i, indx_s+1, sz(vstems)) {
		stem t = vstems[i];
		if(t.inward_extend(s)){
			ret.PB(i);
		}
	}
	return ret;
}

vector<int> stemlist::getstemindice_enclosedby_or_inward_extend(int indx_s){
	vector<int> ret;
	stem s = vstems[indx_s];
	FOR(i, indx_s+1, sz(vstems)) {
		stem t = vstems[i];
		if(s.enclose(t) || t.inward_extend(s)){
			ret.PB(i);
		}
	}
	return ret;
}

//..............((((((((.......)))))))) stem s 
//..(((((.....))))))).................. this stem
vector<int> stemlist::getstemindice_prevparallel_extend(int indx_s) {
	vector<int> ret;
	stem s = vstems[indx_s];
	FOR(i, 0, indx_s-1) {
		stem t = vstems[i];
		if(t.prevparallel_extend(s)) {
			ret.PB(i);
		}
	}
	return ret;
}

//sibling
//.................((((((((.......)))))))) stem s 
//..(((((.....))))).................. this stem
vector<int> stemlist::getstemindice_prevsibling_or_prevparallel_extend(int indx_s) {
	vector<int> ret;
	stem s = vstems[indx_s];
	FOR(i, 0, indx_s-1) {
		stem t = vstems[i];
		if(t.outer.rp == s.outer.lp-1 || t.prevparallel_extend(s)) {
			ret.PB(i);
		}
	}
	return ret;
}

vector<int> stemlist::getstemindice_prevsibling_or_prevparallel_extend(const stem & in_st) {
	vector<int> ret;
	FOR(i, 0, sz(vstems)) {
		stem t = vstems[i];
		if(t.outer.rp == in_st.outer.lp-1 || t.prevparallel_extend(in_st)) {
			ret.PB(i);
		}
		if(t.outer.lp >= in_st.outer.lp) 
			break;
	}
	return ret;
}
	



bool stemlist::conflictwith(stem st) {
	FOR(i,0,sz(vstems)){
		if(vstems[i].conflictwith(st))
			return true;
	}
	return false;
}

bool stemlist::containST(stem st) {
	FOR(i,0,sz(vstems)){
		if(vstems[i] == st) return true;
	}
	return false;
}

//check if the stemlist can form an RNA structure without overlapping stacks, and pseudoknot
bool stemlist::isRNAStructure() {
	int maxlength = 0;
	FOR(i,0,sz(vstems)) {
		maxlength = max(maxlength, vstems[i].outer.rp);
	}
	string str = "";
	FOR(i,0,maxlength) {
		str.PB('.');
	}
	FOR(i,0,sz(vstems)) {
		int x = vstems[i].outer.lp; int y = vstems[i].outer.rp;
		int p = vstems[i].inner.lp; int q = vstems[i].inner.rp;
		FOR(j,x,p+1) {
			if(str[j]!='.') {	
				return false;
			} else {
				str[j] = '('; 
				str[x+y-j] = ')';
			}
		}
	}
	return true;		
}
	

void stemlist::insert(stem & in_st) {
	vstems.insert(vstems.begin(), in_st);
}
void stemlist::insert(stemlist & in_stl) {
	vstems.insert(vstems.begin(), in_stl.vstems.begin(), in_stl.vstems.end());
}
void stemlist::PB(stem & in_st) {
	vstems.PB(in_st);
}
void stemlist::PB(stemlist & in_stl) {
	vstems.insert(vstems.end(), in_stl.vstems.begin(), in_stl.vstems.end());
}

stem & stemlist::operator [] (const int & i) {
	return vstems[i];
}

void stemlist::add(stem & in_st) {
	if(sz(vstems)==0) {
		vstems.PB(in_st);
		return;
	}
	int i = 0; int j = sz(vstems)-1;
	while(i<=j) {
		int mid = (i+j)/2;
		stem m = vstems[mid];
		assert(m!=in_st);
		if(i==j) {
			if(m<in_st) {
				vstems.insert(vstems.begin()+i+1, in_st);
			} else {//m>in_st
				vstems.insert(vstems.begin()+i, in_st);
			}
			return ;
		}
		if(m<in_st) {
			i = mid+1;
		} else if(m>in_st) {
			j = mid-1;
		}
	}
	assert(i>j && i==j+1);
	vstems.insert(vstems.begin()+i, in_st);
}

//return the right-most stem to the left of in_st
//note that, unlike findLS_4FM(), stems in the stemlist do not need to be sorted.
stem stemlist::findLS(const stem & in_st) {
	stem ret;
	FOR(i,0, sz(vstems)) {
		stem sti = vstems[i];
		if(sti < in_st) {
			if(!ret.valid() || (ret.valid() && ret < sti)) 
				ret = sti;
		}
	}
	return ret; 
}

//return the left-most stem to the right of in_st
//note that, unlike findLS_4FM(), stems in the stemlist do not need to be sorted.
stem stemlist::findRS(const stem & in_st) {
	stem ret; 
	FOR(i, 0, sz(vstems)) {
		stem sti = vstems[i];
		if(sti > in_st) {
			if(!ret.valid() || (ret.valid() && sti < ret))
				ret = sti;
		}
	}
	return ret;	
}


//return the right-most stem to the left of in_sg
//note that the stemlist should be able to form a secondary structure (no overlapping stems, no pseudoknot)
stem stemlist::findLS(const segment & in_sg) {
	assert(isRNAStructure());
	stem cover; stem parallel;//to the left
	int lp = in_sg.lp; int rp = in_sg.rp;
	FOR(i, 0, sz(vstems)) {
		if(vstems[i].outer.rp < lp) {
			//vstems[i] to the left of the in_sg segment
			if(!parallel.valid() || (parallel.valid() && vstems[i].outer.rp > parallel.outer.rp)){
				parallel = vstems[i];
			}
		}
		if(vstems[i].inner.lp < lp && rp < vstems[i].inner.rp) {
			//vstems[i] cover the in_sg segment
			if(!cover.valid() || (cover.valid() && vstems[i].inner.lp > cover.inner.lp)) {
				cover = vstems[i];
			}
		}
	}
	
	if(!cover.valid() && parallel.valid())
		return parallel;
	else if(!parallel.valid() && cover.valid()) {	
		return cover; 
	} else if(cover.valid() && parallel.valid()) {
		if(cover.enclose(parallel)) {
			return parallel;
		} else {
			return cover;
		}
	} else //both are invliad
		return stem(-1,-1,-1,-1);
}

//return the left-most stem to the right of in_sg
//note that the stemlist should be able to form a secondary structure (no overlapping stems, no pseudoknot)
stem stemlist::findRS(const segment & in_sg) {
	assert(isRNAStructure());
	stem cover; stem parallel;
	int lp = in_sg.lp; int rp = in_sg.rp;
	FOR(i, 0, sz(vstems)) {
		if(vstems[i].outer.lp > rp) {
			//vstems[i] to the right of the in_sg segment
			if(!parallel.valid() || (parallel.valid() && vstems[i] < parallel)) {
				parallel = vstems[i];
			}
		}
		if(vstems[i].inner.lp < lp && rp < vstems[i].inner.rp) {
			//vstems[i] cover the in_sg segment
			if(!cover.valid() || (cover.valid() && vstems[i] > cover)) {
				cover = vstems[i];
			}
		}
	}
	if(!cover.valid() && parallel.valid())
		return parallel;
	else if(!parallel.valid() && cover.valid()) 
		return cover;
	else if(cover.valid() && parallel.valid()) {
		if(cover.enclose(parallel))
			return parallel;
		else return cover;
	} else {//both are invliad 
		return stem(-1,-1,-1,-1);
	}
}
	

//Attention: make sure that stems in the stemlist are well sorted, otherwise, we can not use binary search in findLS_4FM
//Attention: findLS_4FM can only be used to search the right-most stem to the left of a stem enclosed in a FM segment, for example, in the following, findLS_4FM can only be used when stem 0 and stem 2 are in stemlist and stem 1 is not in stemlist and we want to identify the right-most left stem for stem 1. In this case, return stem 0
// 0    1     1        2     2    0 
//(((..(((...)))......(((...)))..)))
//ALWAYS use stemlist::add(stem) to add a stem to the list
stem stemlist::findLS_4FM(const stem & in_st) {
	int indx = -1;

	if(sz(vstems)!=0) {
		int i = 0; int j = sz(vstems)-1;
		while(i<=j) {
			int mid = (i+j)/2;
			stem m = vstems[mid];
			if(i==j) {
				if(m<in_st) {
					indx = mid;
				} else if(m>in_st) {
					indx = mid-1;
				} else if(m==in_st) {
					indx = mid-1;
				}
				break;
			}
			if(m<in_st) {
				i = mid+1;
			} else if(m>in_st) {
				j = mid-1;
			} else if(m==in_st) {
				indx = mid-1;
				break;
			}
		}
		if(i>j) indx = j;
	}
	
	if(indx<0 || indx>sz(vstems))
		return stem(-1,-1,-1,-1);
	else 
		return vstems[indx];
}

//Attention: make sure that stems in the stemlist are well sorted, otherwise, we can not use binary search in findRS_4FM. 
//Attention: findRS_4FM can only be used to search the left-most stem to the right of a stem enclosed in a FM segment, for example, in the following, findRS_4FM can only be used when stem 0 and stem 2 are in stemlist and stem 1 is not in stemlist and we want to identify the left-most stem to the right of stem 1. In this case, return stem 2
// 0    1     1        2     2    0 
//(((..(((...)))......(((...)))..)))
//ALWAYS use stemlist::add(stem) to add a stem to the list
stem stemlist::findRS_4FM(const stem & in_st) {
	int indx = -1;	
	if(sz(vstems)!=0) {
		int i = 0; int j = sz(vstems)-1;
		while(i<=j) {
			int mid = (i+j)/2;
			stem m = vstems[mid];
			if(i==j) {
				if(m<in_st) {
					indx = mid+1;
				} else if(m>in_st) {
					indx = mid;
				} else if(m==in_st) {
					indx = mid+1;
				}
				break;
			}
			if(m<in_st) {
				i = mid+1;
			} else if(m>in_st) {
				j = mid-1;
			} else if(m==in_st) {
				indx = mid-1;
				break;
			}
		}
		if(i>j) indx = i;
	}
	
	if(indx<0 || indx>sz(vstems))
		return stem(-1,-1,-1,-1);
	else 
		return vstems[indx];
}

/* Given the starting and ending positions, return Fset, a set of indices of stems in vstems.
	For each positive integer x in Fset, let s = vstems[x], then 
		(1) s <_I [i, j] and 
		(2) there should not exist any stem t, t<_I [i,j] and 
			(2.1) s <_P t or 
			(2.2) s <_I t
*/
vector<int> stemlist::getFSet(const int & int_s, const int & in_e) {
	vector<int> ret; 
	//not implemented yet.
	return ret; 
}

stem stemlist::getsubstem_startfrom_endat(const int & i, const int & j) {
	stem ret(-1,-1,-1,-1);
	FOR(k,0,sz(vstems)) {
		stem stk = vstems[k];
		int kx = stk.outer.lp; int ky = stk.outer.rp;
		int kp = stk.inner.lp; int kq = stk.inner.rp;
		if(kx <= i && ky >= j && kp >= i && kq <= j && i+j==kx+ky && kp-i+1>=2) {
			ret = stem(i, j, kp, kq);
		} //else do nothing
	}
	return ret;
}

vector<stem> stemlist::getsubstems_enclosedby_or_inward_extend(const stem & s) {
	vector<stem> ret;
	int sx = s.outer.lp; int sy = s.outer.rp;
	int sp = s.inner.lp; int sq = s.inner.rp;
	
	FOR(k, 0, sz(vstems)) {
		stem stk = vstems[k];
		int kx = stk.outer.lp; int ky = stk.outer.rp;
		int kp = stk.inner.lp; int kq = stk.inner.rp;
		
		if(kx > sy) break;
		if(stk.enclosedby(s)) {
			ret.PB(stk);
		}
		if(stk.inward_extend(s)) {
			if(sx<kx && kx<=sp && sp<kp && ky<sq) {	
				if(kp-sp>=2) ret.PB(stem(sp+1, kx+ky-sp-1, kp, kq));
			} else if(sp<kx && kq<sq && sq<=ky && ky<sy) {
				if(sq-kq>=2) ret.PB(stem(kx+ky-sq+1, sq-1, kp, kq));
			} else if(sx<kx && kx<=sp && sp<kp && kq<sq && sq<=ky && ky<sy) {
				if(kx+ky < sx+sy) {
					if(kp-sp>=2) ret.PB(stem(sp+1, kx+ky-sp-1, kp, kq));
				} else {
					if(sq-kq>=2) ret.PB(stem(kx+ky-sq+1, sq-1, kp, kq));
				}
			} else assert(false);
		}
	}
	return ret;
}

vector<stem> stemlist::getsubstems_between(stem & v, stem & w) {
	assert(v.valid() && w.valid() && w.enclosedby(v)) ;
	int wx = w.outer.lp;
	vector<stem> ret;
	vector<stem> vst = getsubstems_enclosedby_or_inward_extend(v);
	FOR(k,0,sz(vst)) {
		stem stk = vst[k];
		int kx = stk.outer.lp; int ky = stk.outer.rp;
		int kp = stk.inner.lp; int kq = stk.inner.rp;
		assert(stk.enclosedby(v) || stk.inward_extend(v));
		if(ky < wx) ret.PB(stk);
		if(stk.prevparallel_extend(w)) 
			ret.PB(stem(kx+ky-wx+1, wx-1, kp, kq));
	}	
	return ret;
}

vector<stem> stemlist::getsubstems_within(const int & i, const int & j) {
	vector<stem> ret;
//	int start = max(0, i-cf.MAXPARALLELOVERLAP);
//	int end = min(L-1, j+cf.MAXPARALLELOVERLAP);

	int start = max(0, i-cf.MAXOVERLAP);
	int end = min(L-1, j+cf.MAXOVERLAP);
	vector<stem> vst = getstems_within(start, end);
	FOR(k, 0, sz(vst)) {
		stem stk = vst[k];
		int kx = stk.outer.lp; int ky = stk.outer.rp;
		int kp = stk.inner.lp; int kq = stk.inner.rp;
		int x, y, p, q;
		if(i<=kx && ky<=j) {
			ret.PB(stk);
		} else if(kp<i || kq>j) {
			//do nothing
		} else if(kx<i && ky<=j) {
			if(kp-i+1>=2) ret.PB(stem(i, kx+ky-i, kp, kq));
		} else if(i<=kx && y<ky) {
			if(j-kq+1>=2) ret.PB(stem(kx+ky-j, j, kp, kq));
		} else {
			if(i+j>kx+ky) {
				if(kp-i+1>=2) ret.PB(stem(i, kx+ky-i, kp, kq));
			} else {
				if(j-kq+1>=2) ret.PB(stem(kx+ky-j, j, kp, kq));
			}
		}
	}
	return ret;
}

vector<stem> stemlist::getsubstems_startfrom_within(const int & i, const int & j) {
	vector<stem> ret;
	vector<stem> vst = getsubstems_within(i, j);
	FOR(k,0,sz(vst)) {
		stem stk = vst[k];
		if(stk.outer.lp==i) {
			ret.PB(stk);
		}
	}
	return ret;
}

vector<stem> stemlist::getsubstems_endat_within(const int & i, const int & j) {
	vector<stem> ret;
	vector<stem> vst = getsubstems_within(i, j);
	FOR(k,0,sz(vst)) {
		stem stk = vst[k];
		if(stk.outer.rp==j) {
			ret.PB(stk);
		}
	}
	return ret;
}

vector<stem> stemlist::getsubstems_endat(const int & j) {
	return getsubstems_endat_within(0, j);
}


//in which stack is the input base pair.
stem stemlist::inwhichST(const int & in_lp, const int & in_rp) {
	assert(in_lp < in_rp);
	FOR(k,0,sz(vstems)) {
		stem stk = vstems[k];
		int kx = stk.outer.lp; int ky = stk.outer.rp;
		int kp = stk.inner.lp; int kq = stk.inner.rp;
		
		if(kx<=in_lp && in_lp<=kp && kq<=in_rp && in_rp<=ky && in_lp+in_rp==kx+ky) {
			return stk;
		}
	}
	return stem(-1,-1,-1,-1);
}

/***********************************************************
                         seglist
***********************************************************/
segment seglist::findRS(const segment & in_sg) {
	int indx = -1;
	cout<<"\tcall findRS("<<in_sg.lp<<","<<in_sg.rp<<") from seglist:"<<endl;
	show();
	cout<<endl;
	
	if(sz(vsegs)!=0) {
		int i = 0; int j = sz(vsegs)-1;
		while(i<=j) {
			int mid = (i+j)/2;
			segment m = vsegs[mid];
			if(i==j) {
				if(m<in_sg) {
					indx = mid+1;
				} else if(m>in_sg) {
					indx = mid;
				} else if(m==in_sg) {
					indx = mid+1;
				}
				break;
			}
			if(m<in_sg) {
				i = mid+1;
			} else if(m>in_sg) {
				j = mid-1;
			} else if(m==in_sg) {
				indx = mid-1;
				break;
			}
		}
		if(i>j) indx = i;
	}
	
	if(indx<0 || indx>=sz(vsegs))
		return segment(-1,-1);
	else 
		return vsegs[indx];

}

/***********************************************************
                         seglistT
***********************************************************/
segmentT seglistT::findRS(const segmentT & in_sg) {
	int indx = -1;
	cout<<"\tcall findRS("<<in_sg.lp<<","<<in_sg.rp<<") from seglist:"<<endl;
	show();
	cout<<endl;
	
	if(sz(vsegs)!=0) {
		int i = 0; int j = sz(vsegs)-1;
		while(i<=j) {
			int mid = (i+j)/2;
			segmentT m = vsegs[mid];
			if(i==j) {
				if(m<in_sg) {
					indx = mid+1;
				} else if(m>in_sg) {
					indx = mid;
				} else if(m==in_sg) {
					indx = mid+1;
				}
				break;
			}
			if(m<in_sg) {
				i = mid+1;
			} else if(m>in_sg) {
				j = mid-1;
			} else if(m==in_sg) {
				indx = mid-1;
				break;
			}
		}
		if(i>j) indx = i;
	}
	
	if(indx<0 || indx>=sz(vsegs))
		return segmentT(-1,-1, etype_unknown);
	else 
		return vsegs[indx];

}

/***********************************************************
                         RStack
***********************************************************/
//identify stems in str (the secondary structure) and construct an RStack obj
RStack::RStack(const string & str) {
	L = sz(str);
	vector<BP> vbp = s2v(str);
	
	if (sz(vbp)==0) return;
	sort(vbp.begin(), vbp.end(), ascendingsortBP);

	BP outer = vbp[0];
	BP inner = vbp[0];
	int i=1;
	while(i<sz(vbp)) {
		BP bpx = vbp[i-1];
		BP bpy = vbp[i];
		if (bpx.lp==bpy.lp-1 && bpx.rp==bpy.rp+1) {
			//continue to read the stack
		} else {
			vstems.PB(stem(outer, inner));
			outer = bpy;
		}
		inner = bpy;
		i++;
	}
	vstems.PB(stem(outer, inner));
	sort(vstems.begin(), vstems.end(), ascendingsortST);
}

string RStack::tostring(){
	string ret;
	FOR(i,0,L) ret += ".";
	FOR(i,0,sz(vstems)){
//l1<=l2<r2<r1
		int l1 = vstems[i].outer.lp;
		int l2 = vstems[i].inner.lp;
		int r1 = vstems[i].outer.rp;
		int r2 = vstems[i].inner.rp;
		FOR(j,l1,l2+1){
			ret[j] = '('; ret[l1+r1-j] = ')';
		}
	}
	return ret;
}
//check whether this instance contains stack st?
bool RStack::containST(const stem & st) {
	FOR(i,0,sz(vstems)) {
		if(vstems[i] == st) return true;
	}
	return false;
}

int RStack::inwhichST(const int & query) { 
	//in which stack is query.
	FOR(i,0,sz(vstems)) {
		stem st = vstems[i];
		int x = st.outer.lp; int y = st.outer.rp;
		int p = st.inner.lp; int q = st.inner.rp;
		if((x<=query && query<=p) || (q<=query && query<=y)) 
			return i;
	}
	return -1;
}

//isLocOpt checks whether this instance is local minimum or not.
bool RStack::isLocOpt(const stemlist & in_stl) {
	stemlist rs_stl;
	rs_stl.vstems = vstems;
	FOR(i,0,sz(in_stl.vstems)){
		stem sti = in_stl.vstems[i];
		if(containST(sti)) continue;
		if(!rs_stl.conflictwith(sti)){
			return false;
		}
	}
	return true;
}

vector<RStack> RStack::getneighbor_delstack() {
	vector<RStack> ret;
	FOR(i,0,sz(vstems)) {
		vector<stem> vst;
		FOR(j,0,sz(vstems)) {
			if(i==j) continue;
			vst.PB(vstems[j]);
		}
		RStack tmp;
		tmp.vstems = vst;
		ret.PB(tmp);
	}
	return ret;
}

bool RStack::ismatch(vector<vector<bool>> & ali_pi) {
	FOR(i,0,sz(vstems)) {
		stem s = vstems[i];
		int sx = s.outer.lp; int sy = s.outer.rp;
		int sp = s.inner.lp; int sq = s.inner.rp;
		FOR(k,sx,sp+1) {
			if(!ali_pi[k][sx+sy-k]) {
				cerr<<"msa["<<k<<"]["<<sx+sy-k<<" does not pair."<<endl;
				return false;
			}
		}
	}
	return true;
}

vector<int> RStack::findConflictingST(stem & st) {
	//return indices of stems that conflict with sti
	vector<int> ret;
	FOR(i,0,sz(vstems)) {
		if(st.conflictwith(vstems[i])) {
			ret.PB(i);
		}
	}
	return ret; 
}
/***********************************************************
                     END OF RStack
***********************************************************/


/******************************************************************************
                              PartStackStruct
 ******************************************************************************/
bool PartStackStruct::isLocOpt(stemlist & in_stl) {
	FOR(i,0,sz(in_stl.vstems)){
		stem sti = in_stl.vstems[i];
		int sx = sti.outer.lp; int sy = sti.outer.rp;
		if(stl.containST(sti)) continue;
		if(!stl.conflictwith(sti) && !sgl.overlap(sti) ) {
//			cout<<"The structure is not local optimal because the stem ("<<sti.outer.lp<<", "<<sti.outer.rp<<") ~ ("<<sti.inner.lp<<", "<<sti.inner.rp<<") neither conflicts with any element in the stemlist, nor overlaps with any segment in sgl."<<endl;
			return false;
		}
	}
	return true;
}

/***********************************************************
                         Common
***********************************************************/
//return the index of query if query is in vi
//otherwise return -1
//if direction=1, search in [startpos...sz(vi)-1], increasing order
//if direction=-1, search in from [startpos..0], decreasing order
int getindx(const vector<int> & vi, const int & query, const int & startpos, const int & direction){
	if(direction==1) {
		int i = (startpos>=0)?(startpos):0;
		for(; i<sz(vi); i++) {
			if(vi[i]==query) return i;
		}
		return -1;
	} else if(direction==-1) {
		if(startpos<0) return -1;
		int i = (startpos>=sz(vi)-1)?(sz(vi)-1):(startpos);
		for(; i>=0; i--) {
			if(vi[i]==query) return i;
		}
		return -1;
	} else {
		cerr<<"direction can either be 1 or -1 while it is "<<direction<<endl; exit(0);
	}
}

//return the index of in_bp in in_vbp
//if direction=1, search in [startpos...sz(in_vbp)-1], increasing order
//if direction=-1, search in from [startpos..0], decreasing order
int getindx(const vector<BP> & in_vbp, const BP & in_bp, const int & startpos, const int & direction) {
	if(direction==1) {
		int i = (startpos>=0)?(startpos):0;
		for(; i<sz(in_vbp); i++) {
			if(in_vbp[i].lp == in_bp.lp && in_vbp[i].rp==in_bp.rp) return i;
		}
		return -1;
	} else if(direction==-1) {
		if(startpos<0) return -1;
		int i = (startpos>=sz(in_vbp)-1)?(sz(in_vbp)-1):(startpos);
		for(; i>=0; i--) {
			if(in_vbp[i].lp == in_bp.lp && in_vbp[i].rp==in_bp.rp) return i;
		}
		return -1;
	} else {
		cerr<<"direction can either be 1 or -1 while it is "<<direction<<endl; exit(0);
	}
}


int getindx(const vector<stem> & in_vst, const stem & in_st, const int & startpos, const int & direction) {
	stem myst = in_st;
	if(direction==1) {
		int i = (startpos>=0)?(startpos):0;
		for(;i<sz(in_vst);i++){
			if(myst == in_vst[i]) return i;
		}
		return -1;
	} else if(direction==-1){
		if(startpos<0) return -1;
		int i = (startpos>=sz(in_vst)-1)?(sz(in_vst)-1):(startpos);
		for(; i>=0; i--){
			if(myst == in_vst[i]) return i;
		}
		return -1;
	} else {
		cerr<<"direction can either be 1 or -1 while it is "<<direction<<endl; exit(0);
	}
}

//given a string representation of a strucuture, return vbp
vector<BP> s2v(const string & str)
{
#ifdef DEBUG
	cout<<"in s2v, str="<<str<<", sz(str)="<<sz(str)<<endl;
#endif
	stack<int> sta;
	vector<BP> ret;
	FOR(i,0,sz(str)) {
#ifdef DEBUG
//		cout<<"i="<<i<<", s[i]="<<s[i]<<endl;
#endif
		if (str[i] == '(') {
			sta.push(i);
		} else if(str[i] == ')') {
			assert(sz(sta)!=0);
			ret.PB(BP(sta.top(), i));
			sta.pop();
		}
#ifdef DEBUG
//		cout<<"the size of stack is "<<sz(sta)<<endl;
#endif
	}
	assert(sz(sta)==0);
	return ret;
}

//given indice of the left/right parenthesis and the length of the RStruct, return s
string v2s(const int & l, const vector<BP> & in_vbp)
{
#ifdef DEBUG
//	cout<<"in v2s, len="<<l<<endl;
#endif 

	string str = "";	
	FOR(i,0,l)
		str += " ";
	
	FOR(i,0,sz(in_vbp)) {
		str[in_vbp[i].lp] = '(';
		str[in_vbp[i].rp] = ')';
	}
	return str;
}


//return indices of unpaired bases
vector<int> getunp(const int & l, const vector<BP> & in_vbp) {
	vector<int> ret; 
	
	int tag_unp[l]; //tag_unp[i] represents if s[i] is an unpaired base
	FOR(i,0,l) {
		tag_unp[i] = 1; //assume every index is in unp
	}
	
	FOR(i,0,sz(in_vbp)) {
		tag_unp[in_vbp[i].lp] = 0;
		tag_unp[in_vbp[i].rp] = 0;
	}
	
	FOR(i,0,l) 
		if(tag_unp[i]==1)
			ret.PB(i);

	return ret;
}

//return indices of unpaired bases
vector<int> getunp(const string & str) {
	vector<int> ret; 
	FOR(i,0,sz(str)) 
		if(str[i]=='(' ||  str[i]==')'){}
		else	ret.PB(i);
	return ret;
}



bool descendingsort(double d1, double d2) {
	return (d1 > d2);
}
bool descendingsortBP(BP bp1, BP bp2) {
	return (bp1.lp > bp2.lp);
}
bool descendingsortST(stem st1, stem st2) {
	if (st1.outer.lp == st2.outer.lp) {
		return (st1.outer.rp > st2.outer.rp);
		//why not compare (st1.outer.rp == st2.outer.rp)? Because any two stacks \
		can not share exactly the same outer.lp and outer.rp
	} else
		return (st1.outer.lp > st2.outer.lp);
}

bool ascendingsort(double d1, double d2) {
	return (d1 < d2);
}
bool ascendingsortBP(BP bp1, BP bp2) {
	return (bp1.lp < bp2. lp);
}
bool ascendingsortST(stem st1, stem st2) {
	if (st1.outer.lp == st2.outer.lp) {
		return (st1.outer.rp < st2.outer.rp);
		//why not compare (st1.outer.rp == st2.outer.rp)? Because any two stacks \
		can not share exactly the same outer.lp and outer.rp
	} else
		return (st1.outer.lp < st2.outer.lp);
}

//check if in_bp clash with any base pair in in_vbp
bool mycheckclash(const BP & in_bp, const vector<BP>& in_vbp) {
	int x=in_bp.lp; int y=in_bp.rp;
	
	FOR(k,0,sz(in_vbp)) {
		int i = in_vbp[k].lp; int j = in_vbp[k].rp;
		//clash A: i ... x ... j ... y
		//clash B; x ... i ... y ... j
		if (i < x && x < j && j < y) return true;
		if (x < i && i < y && y < j) return true;
		if (x==i || x==j || y==i || y==j) return true;
	}
	return false;	
}


