#include "Turner.h"

string trimspace(const string & str) {
	int s = 0,  e = sz(str)-1; 
	bool stop = false;
	FOR(i,0,sz(str)) {
		if(str[i] == ' ' || str[i] == '\t' ||  str[i] == '\n') {
		} else {
			s = i;
			stop = true;
			break;
		}
	}
	if(!stop) return "";
	
	for(int i= sz(str)-1; i>=s; i--) {
		if(str[i] == ' ' || str[i] == '\t' ||  str[i] == '\n') {
		} else {
			e = i;
			break;
		}
	}
	string ret = str.substr(s, (e-s+1));
	return ret;
}

void TurnerE::init(const string & LEfile, const string & SEfile, const string & 
  TerminalSMHEfile, const string & TerminalSMIEfile, const string & 
  SingleMIEfile, const string & OneTwoMIEfile, const string & TetraLoopEfile, 
  const string & TwoTwoMIEfile, const string & d3Efile, const string & d5Efile) {
	init_LE(LEfile);
	init_SE(SEfile);
	init_TerminalSMHE(TerminalSMHEfile);
	init_TerminalSMIE(TerminalSMIEfile);
	init_SingleMIE(SingleMIEfile);
	init_OneTwoMIE(OneTwoMIEfile);
	init_TetraLoopE(TetraLoopEfile);
	init_TwoTwoMIE(TwoTwoMIEfile);
	init_d3E(d3Efile);
	init_d5E(d5Efile);
}

void TurnerE::init_LE(const string & file)
{
//	cout<<"LoopEfile="<<file<<endl;
//format: #Size INTERNAL BULGE	HAIRPIN                                       \
1	.	3.80	.                                                                  \
2	.	2.80	.                                                                  \
3	.	3.20	5.70                                                               \
4	1.70	3.60	5.60                                                            \
5	1.80	4.00	5.60                                                            \
6	2.00	4.40	5.40                                                            \
7	2.20	4.60	5.90

	ifstream fin(file.c_str());
	while(fin.good() && !fin.eof()){
		char buf[MAXCHARNUM];
		fin.getline(buf, MAXCHARNUM);
		string str(buf);
		str = trimspace(str);
		if(str[0] == '#' || sz(str)==0) {
			//comments, don't process
			continue;
		}
		stringstream ss(str); 
		string s1, s2, s3, s4;
		ss>>s1>>s2>>s3>>s4;
		int i1;
		int i2, i3, i4;
		stringstream ss1(s1); ss1>>i1;
		if(s2.compare(".") != 0) {
			i2 = atoi(s2.c_str());
			INTERNAL[i1] = i2;
		}
		if(s3.compare(".") != 0) {
			i3 = atoi(s3.c_str());
			BULGE[i1] = i3;
		}
		if(s4.compare(".") != 0) {
			i4 = atoi(s4.c_str());
			HAIRPIN[i1] = i4;
		}
	}
	fin.close();
		
	Mh = -100;//energy for every hydrogen bond
	Mb = 10;//penalty for an unpaired nt in a dangling region
	Mc = 340;//penalty for opening a multiloop
	Mphi = 40;//penalty for every branch in a multiloop
	Mbb = 0; //penalty for every unpaired nt in multiploop
	Mninio_m = 50;//penalty for difference of lengths between two gaps of an interior loop
	Mninio_max = 300; //max penalty for difference of lengths between two gaps of an interior loop
	MtermianAU = 50;//penalty for having AU in the terminal
	Mlongloop = 107.9;//For internal/bulge/hairpin loops > 30, dS(T) = ds(30)+Mlongloop*ln(n/30), Mlongloop = 1.079
	Mgubous = -220;
	Mcslope = 30;
	Mcintercept = 160;
	Mc3 = 140;
}

void TurnerE::readfiletype(const string & file, map <string, int> & table, 
  const int & pos1, const int & pos2) {
//format:                                                                      \
#         UX                                                                   \
#         GY                                                                   \
>UXYG                                                                          \
  .     .     .   -100                                                         \
  .     .   -150   .                                                           \
  .   -140   .    30                                                           \
-60     .   -50   .                                                            \
assure name[pos1]=='X' and name[pos2]=='Y'
	char alphabet[4];
	bool setalphabet = false;
	ifstream fin(file.c_str());
	while(fin.good() && !fin.eof()){
		char buf[MAXCHARNUM];
		fin.getline(buf, MAXCHARNUM);
		string str(buf);
		str =  trimspace(str);
		if(sz(str)==0 || str[0] == '#') {
			//comments, don't process
			continue;
		}
		stringstream ss(str);
		string s1, s2;
		if(str.find("alphabet") != string::npos) {
			str = trimspace(str.substr(str.find("=")+1));
			FOR(i,0,4)
				alphabet[i] = str[i];
			setalphabet=true;
			continue;
		}
		string name;
		vector<string> vval;
		FOR(i,0,4) vval.PB("");
		
		if(str[0] == '>') {
			if(!setalphabet) {
				cerr<<"file:"<<file<<" does not specify alphabet, please add alphabet=ACGU to the first line."<<endl;
				exit(0);
			}
			name = str.substr(1);
			if(pos2>=0)	assert(name[pos1]='X' && name[pos2]=='Y');
			else assert(name[pos1]=='X');
			if(pos2>=0) {
				FOR(i,0,4) {			
					//row i
					fin.getline(buf, MAXCHARNUM);
					string str(buf);
					str = trimspace(str);
					stringstream ss2(str);
					FOR(j,0,4) {
						//column j
						ss2>>vval[j];
						if(vval[j].compare(".")!=0) {
							string newname = name;
							newname[pos1] = alphabet[i];
							newname[pos2] = alphabet[j];
							stringstream ss3(vval[j]);
							int ival;
							ss3>>ival;
							table[newname] = ival;
						}
					}
				}
			} else {
				fin.getline(buf, MAXCHARNUM);
				string str(buf);
				str = trimspace(str);
				stringstream ss2(str);
				FOR(j,0,4) {
					//colum j;
					string newname = name;
					newname[pos1] = alphabet[j];
					int ival;
					ss2 >> ival;
					table[newname] = ival;
				}
			}
		}
	}
}

void TurnerE::init_SE(const string & file) {
	readfiletype(file, STACKING, 1, 2);
}
void TurnerE::init_TerminalSMHE(const string & file) {
	readfiletype(file, TerminalSMHE, 1, 2);
}
void TurnerE::init_TerminalSMIE(const string & file) {
	readfiletype(file, TerminalSMIE, 1, 2);
}
void TurnerE::init_SingleMIE(const string & file) {
	readfiletype(file, SingleMIE, 1, 4);
}
void TurnerE::init_OneTwoMIE(const string & file) {
	readfiletype(file, OneTwoMIE, 1, 5);
}

void TurnerE::init_TetraLoopE(const string & file) {
	ifstream fin(file.c_str());
	while(fin.good() && !fin.eof()){
		char buf[MAXCHARNUM];
		fin.getline(buf, MAXCHARNUM);
		string str(buf);
		str =  trimspace(str);
		if(sz(str)==0 || str[0] == '#') {
			//comments, don't process
			continue;
		}
		stringstream ss(str);
		string s1, s2;
		ss>>s1>>s2;
		s1 = trimspace(s1);
		assert(sz(s1)==6);
		int ival;
		stringstream ss2(s2);
		ss2>>ival;
		TetraLoopE[s1] = ival;
	}
	fin.close();
}

void TurnerE::init_TwoTwoMIE(const string & file) {
	char alphabet[4];//by default, ACGU
	vector<string> alphabet2;//by default, AA AC AG AU CA CC CG CU GA GC GG GU UA UC UG UU
	bool setalphabet = false;
//format: 
//>AXYAUQPU
	int posX = 1; int posY = 2; int posP = 6; int posQ = 5;
	ifstream fin(file.c_str());
	while(fin.good() && !fin.eof()){
		char buf[MAXCHARNUM];
		fin.getline(buf, MAXCHARNUM);
		string str(buf);
		str =  trimspace(str);
		if(sz(str)==0 || str[0] == '#') {
			//comments, don't process
			continue;
		}
		stringstream ss(str);
		string s1, s2;
		if(str.find("alphabet") != string::npos) {
			str = trimspace(str.substr(str.find("=")+1));
			FOR(i,0,4)
				alphabet[i] = str[i];
			FOR(i,0,4) FOR(j,0,4) {
				string tmp = ""; tmp += alphabet[i]; tmp += alphabet[j];
				alphabet2.PB(tmp);
			}
			setalphabet=true;
			continue;
		}

		string name;
		vector<string> vval;
		FOR(i,0,sz(alphabet2)) vval.PB("");
		
		if(str[0] == '>') {
			if(!setalphabet) {
				cerr<<"file:"<<file<<" does not specify alphabet, please add alphabet=ACGU to the first line."<<endl;
				exit(0);
			}
			name = str.substr(1);
			assert(name[posX]=='X' && name[posY]=='Y' && name[posP] == 'P' && name[posQ] == 'Q');
		}		
		FOR(i,0,sz(alphabet2)) {
			//row i, alphabet2[i] = (XP)
			fin.getline(buf, MAXCHARNUM);
			string str(buf);
			str = trimspace(str);
			stringstream ss2(str);
			FOR(j,0,sz(alphabet2)) {
				ss2>>vval[j];
				if(vval[j].compare(".")!=0) {
					string newname = name;
					newname[posX] = alphabet2[i][0];
					newname[posP] = alphabet2[i][1];
					newname[posY] = alphabet2[j][0];
					newname[posQ] = alphabet2[j][1];
					stringstream ss3(vval[j]);
					int ival; ss3>>ival;
					TwoTwoMIE[newname] = ival;
				}
			}
		}
	}
}


void TurnerE::init_d3E(const string & file) {
	readfiletype(file, d3E, 1, -1);
}

void TurnerE::init_d5E(const string & file) {
	readfiletype(file, d5E, 1, -1);
}
















