#ifndef _TURNER_H_
#define _TURNER_H_
#include <map>
#include <math.h>
#include "Util.h"
#include "TurnerParam.h"

#define MAXLOOPLEN 30 //the length of the longest loop in a bulge, an interior \
loop or a hairpin loop
//#define MAXINTERIORLENDIFF 10//the difference between lengths of two loops in  \
an interior loop

class TurnerE{
public:
	map<int, int> INTERNAL;//Destablizing energy of interior loops
	map<int, int> BULGE;//Destablizing energy of bulges
	map<int, int> HAIRPIN;//Destablizing energy of hairpin loops
	
	map<string, int> STACKING;//stacking energy
	map<string, int> TerminalSMHE;//terminal stacking mismatch hairpin energy
	map<string, int> TerminalSMIE;//terminal stacking mismatch interior energy
	map<string, int> SingleMIE;//single mismatch interior energy
	map<string, int> OneTwoMIE;//1x2 mismatch interior energy
	map<string, int> TetraLoopE;//specially stable tetra loop energy;
	map<string, int> TwoTwoMIE;//2x2 mismatch interior energy
	
	map<string, int> d3E;//dangling single base stacking energy at 3' end
	map<string, int> d5E;//dangling single base stacking energy at 5' end
	
	int Mh;//energy for every hydrogen bond
	int Mb;//penalty for an unpaired nt in a dangling region
	int Mc;//penalty for opening a multiloop
	int Mphi;//penalty for having every branch in a multiloop
	int Mbb;//panlaty for an unpaired nt in a multiloop
	int Mninio_m;//penalty for difference of lengths between two gaps of an interior loop
	int Mninio_max; //max penalty for difference of lengths between two gaps of an interior loop
	int MtermianAU;//penalty for having AU in the terminal;
	double Mlongloop; //For internal/bulge/hairpin loops > 30, dS(T) = ds(30)+Mlongloop*ln(n/30), Mlongloop = 1.079
	int Mgubous; //For GU enclosure preceded by GG
	int Mcslope; //c hairpin slope 
	int Mcintercept; //c hairpin intercept 
	int Mc3; //c hairpin of 3
	
	TurnerE(){}
	TurnerE(const string & LEfile, const string & SEfile, const string & 
	  TerminalSMHEfile, const string & TerminalSMIEfile, const string & 
	  SingleMIEfile, const string & OneTwoMIEfile, const string & TetraLoopEfile,
	  const string & TwoTwoMIEfile, const string & d3Efile, const string & d5Efile) {
		init(LEfile, SEfile, TerminalSMHEfile, TerminalSMIEfile, SingleMIEfile, OneTwoMIEfile, TetraLoopEfile, TwoTwoMIEfile, d3Efile, d5Efile);
	}
	void init() {
		init(LEfile, SEfile, TerminalSMHEfile, TerminalSMIEfile, SingleMIEfile, OneTwoMIEfile, TetraLoopEfile, TwoTwoMIEfile, d3Efile, d5Efile);
	}
	void init(const string & LEfile, const string & SEfile, const string & 
	  TerminalSMHEfile, const string & TerminalSMIEfile, const string & 
	  SingleMIEfile, const string & OneTwoMIEfile, const string & TetraLoopEfile,
	  const string & TwoTwoMIEfile, const string & d3Efile, const string & d5Efile);
	void readfiletype(const string & file, map <string, int> & table, 
	  const int & pos1, const int & pos2);
	void init_LE(const string & LEfile);
	void init_SE(const string & SEfile);
	void init_TerminalSMHE(const string & TerminalSMHEfile);
	void init_TerminalSMIE(const string & TerminalSMIEfile);
	void init_SingleMIE(const string & SingleMIEfile);
	void init_OneTwoMIE(const string & OneTwoMIEfile);
	void init_TetraLoopE(const string& TetraLoopEfile);
	void init_TwoTwoMIE(const string & TwoTwoMIEfile);
	void init_d3E(const string & d3Efile);
	void init_d5E(const string & d5Efile);
};

#endif

