#include "clusterRS.h"

bool operator == (const stemstem & stst1, const stemstem & stst2) {
	if(stst1.sti == stst2.sti && stst1.stj == stst2.stj) 
		return true; 
	else return false;
}

bool operator != (const stemstem & stst1, const stemstem & stst2){
	if(stst1.sti != stst2.sti || stst1.stj != stst2.stj) 
		return true; 
	else return false;
}

int StemCompareHelp(const stem & A, const stem & B)
{
	if(A.outer.lp != B.outer.lp) return (A.outer.lp - B.outer.lp);
	if(A.outer.rp != B.outer.rp) return (A.outer.rp - B.outer.rp);
	if(A.inner.lp != B.inner.lp) return (A.inner.lp - B.inner.lp);
	if(A.inner.rp != B.inner.rp) return (A.inner.rp - B.inner.rp);
	return 0;
}

bool operator < (const stemstem & stst1, const stemstem & stst2) {
	stem sti1 = stst1.sti; stem stj1 = stst1.stj;
	stem sti2 = stst2.sti; stem stj2 = stst2.stj;
	
	int c = StemCompareHelp(sti1, sti2);
	if(c != 0) return c < 0;
	return (StemCompareHelp(stj1, stj2) < 0);
}

bool operator > (const stemstem & stst1, const stemstem & stst2) {
	return stst2 < stst1;
}

vector<int> RStack2vectorint(const RStack & rs) {	
	RStack rs1(rs);
	vector<int> ret;
	FOR(i,0,rs.L) ret.PB(-1);
	FOR(i,0,sz(rs.vstems)) {
		stem s = rs.vstems[i];
		int x = s.outer.lp; int y = s.outer.rp; 
		int p = s.inner.lp; int q = s.inner.rp;
		FOR(k,x,p+1) {
			assert(ret[k] == -1 && ret[x+y-k] == -1);
			ret[k] = x+y-k;
			ret[x+y-k] = k;
		}
	}
	return ret;
}

vector<int> str2vectorint(const string & str) {
	return RStack2vectorint(RStack(str));
}

vector<int> compute_conflictingdegree_between_stems(stem & sti, stem & stj) {
//	cout<<"the conflicting degree between "; sti.show(); cout<<" "; stj.show(); cout<<endl;
	int ix = sti.outer.lp; int iy = sti.outer.rp;
	int ip = sti.inner.lp; int iq = sti.inner.rp;

	int jx = stj.outer.lp; int jy = stj.outer.rp;
	int jp = stj.inner.lp; int jq = stj.inner.rp;
	vector<int> ret; 
	//if sti and stj are identical 
	if(sti==stj) {
		FOR(i,0,ip-ix+1) {
			ret.PB(0);
		}
		return ret;
	}

	//case -1 : not set;	
	//case 0  : set to 0 for sure;
	//case 1  : set to 1 for sure;
	//case 2  : set to samedirect or greater;
	FOR(i,0,ip-ix+1) {
	//initiallization, -1 means not set 
		ret.PB(-1);
	}
		
	if(jy<ix || iy<jx || !sti.conflictwith(stj)) {
		//since sti and stj do not conflict with each other, return an array ret of size 0.
		return ret;
	}
		
	FOR(i,ix,ip+1) {
		int k = ix+iy-i;
		int val;
		int segi = stj.inwhichseg(i);
		int segk = stj.inwhichseg(k);
		if( (segi==1 && segk==1) ||
		    (segi==3 && segk==3) ||
		    (segi==5 && segk==5) ||
		    (segi==1 && segk==5) ) {
		    val = -1; //not set;
		} else if( 
		    (segi==1 && segk==2) ||
		    (segi==4 && segk==5) ||
		    (segi==2 && segk==2) ||
		    (segi==4 && segk==4) ||
		    (segi==1 && segk==3) ||
		    (segi==3 && segk==5) ) {
		    val = 1; //set to 1
		} else if(
		    (segi==2 && segk==3) ||
		    (segi==3 && segk==4) ||
		    (segi==1 && segk==4) ||
		    (segi==2 && segk==5) ) {
		    val = 2; //set to samedirect or greater.
		} else if(
		    (segi==2 && segk==4) ) {
		    if(ix+iy == jx+jy) {
		    	val = 0;
		    } else {
			 	val = samedirect;
			 }
		} else assert(false);
		ret[i-ix] = val;
	}
	return ret;
}

vector<int> get_conflictingdegree_between_stems(map<stemstem, vector<int>> & CDSTmap, stem & sti, stem & stj) {
	int length = sti.inner.lp-sti.outer.lp+1;
	stemstem stst(sti, stj);
	map<stemstem, vector<int>>::iterator it;
	it = CDSTmap.find(stst);
	if(it!=CDSTmap.end()) {
		return (*it).second;
	} else {
		vector<int> value = compute_conflictingdegree_between_stems(sti, stj);
		assert(sz(value) == length);
		CDSTmap[stst] = value;
		return value;
	}
}

vector<double> compute_conflictingdegree(RNAConsT & in_rc, const string & s1, 
               const string & s2, map<stemstem, vector<int>> & CDSTmap) {
	RStack rs1(s1);
	RStack rs2(s2);
	return compute_conflictingdegree(in_rc, rs1, rs2, CDSTmap);
}

vector<double> compute_conflictingdegree(RNAConsT & in_rc, RStack & rs1, RStack & rs2, map<stemstem, vector<int>> & CDSTmap) {
	vector<double> ret;
	vector<int> vi1 = RStack2vectorint(rs1);
	vector<int> vi2 = RStack2vectorint(rs2);
	
	FOR(i,0,sz(rs1.vstems)) {
		stem sti = rs1.vstems[i];
		int ix = sti.outer.lp; int iy = sti.outer.rp;
		int ip = sti.inner.lp; int iq = sti.inner.rp;
		int leni = ip-ix+1;
		
		if(rs2.containST(sti)) {
			ret.PB(0);
			continue;
		}
		vector<int> vcdi;
		FOR(k,0,leni) {
/* vcdi[.]= -1: value is not set;
   vcdi[.]= 0 : set to 0 for sure;
   vcdi[.]= 1 : set to 1 for sure;
   vcdi[.]= 2 : set to samedirect or greater; */
			vcdi.PB(-1);
		}

		vector<int> vindx2 = rs2.findConflictingST(sti);
		FOR(indxj,0,sz(vindx2)) {
			int j = vindx2[indxj];
			stem stj = rs2.vstems[j];
			assert(sti.conflictwith(stj));
			vector<int> vcdij = get_conflictingdegree_between_stems(CDSTmap, sti, stj);
			//compute_conflictingdegree_between_stems(sti, stj);
			if(sz(vcdi) != sz(vcdij)) {
				cout<<"\nsti is "<<endl;
				sti.show();
				cout<<"\nstj is "<<endl;
				stj.show();
				cout<<"size of vcdi is "<<sz(vcdi)<<endl;
				cout<<"size of vcdij is "<<sz(vcdij)<<endl;
				cout<<"length of sti is "<<(sti.inner.lp-sti.outer.lp+1)<<endl;
			}
			assert(sz(vcdi) == sz(vcdij));
			bool stop = true;
			FOR(k,0,sz(vcdij)) {
				if(vcdi[k]==-1) {
					vcdi[k] = vcdij[k];
					stop = false;
				} else if(vcdi[k]==0) {
					//set to 0 for sure
					assert(vcdij[k]==-1 || vcdij[k]==0);
				} else if(vcdi[k]==1) {
					assert(vcdij[k]!=0);
				} else if(vcdi[k]==2) {
					assert(vcdij[k]!=0);
					if(vcdij[k]==1) {
						vcdi[k] = 1;
					} else {
						stop = false;
					}
					//else if(vcdij[k]==-1) vcdi[k] = 2;
					//else if(vcdij[k]==2 ) vcdi[k] = 2;
				}
			}
			if(stop) break;
		}

		double sum = 0;
		FOR(k,0,sz(vcdi)) {
			if(vcdi[k]==-1) {
				sum += 1; //if never been set, the position counts up to 1
			} else if(vcdi[k]==0) {
				//if this position counts 0
				//sum += 0;
			} else if(vcdi[k]==1) {
				//if this position counts 1
				sum += 1;
			} else if(vcdi[k]==2) {
				//if this position counts samedirect or greater
				sum += samedirect;
			}
		}
		ret.PB(sum/(double)(leni));
	}
	return ret;
}

//compute stacking energy of all stems in vst and return a map mapping stems to 
//their energies (which are integers);
map<stem, int> computeStemE(RNAConsT & in_rc, vector<stem> & vst) {
	map<stem, int> ret;
	FOR(i,0,sz(vst)) {
		stem sti = vst[i];
		int e = in_rc.computeS(in_rc.msa, sti); //computeS(msa, rs1.vstems[h])
		ret[sti] = e;
	}
	return ret;
}

//obtain the stacking energy of stem st from map stemEmap, if it is not 
//available, call computeS and add it to stemEmap.
int getStemE(RNAConsT & in_rc, stem & st, map<stem, int> & stemEmap) {
	map<stem,int>::iterator it;
	it = stemEmap.find(st);
	if(it!=stemEmap.end()) {
		return (*it).second;
	} else {
		int e = in_rc.computeS(in_rc.msa, st);
		stemEmap[st] = e;
		return e;
	}
}

double compute_pairwise_conflictingdegree(RNAConsT & in_rc, 
      RStack & rs1, RStack & rs2, map<stem, int> & stemEmap, 
      map<stemstem, vector<int>> & CDSTmap) {
//computing the pairwise distance (approximated barrier) between rs1 and rs2.
	vector<double> vd = compute_conflictingdegree(in_rc, rs1, rs2, CDSTmap);
	double sum = 0;
	FOR(h,0,sz(vd)) {
		int e = getStemE(in_rc, rs1.vstems[h], stemEmap);
		sum += (vd[h] * (double)e)/(double)UNIT/(double)in_rc.N;
	}
	return sum ;
}
