#include "Config.h"
/***********************************************************
                         Config
***********************************************************/
void Config::setConfig(const string & configfile) {
	if(!fexists(configfile.c_str())) {
		cerr<<"The specified config file does not exist."<<endl;
		exit(1);
	}
	ifstream configin(configfile.c_str());
	char buf[MAXCHARNUM];
	while(!configin.eof()){
		configin.getline(buf, MAXCHARNUM);
		string sline = string(buf);
		if(sline[0] == '#') continue;
		if(sz(sline)==0) continue;
		int indx = sline.find("=");
		if(indx != string::npos) {
			string name = sline.substr(0,indx);
			string val = sline.substr(indx+1);
			name = trim(name);
			val = trim(val);
			istringstream iss(val);
			if(name.compare("ALLOWGU")==0){
				if(val.compare("TRUE")==0 || val.compare("True")==0 || val.compare("true")==0) {
					ALLOWGU = true;
				} else if(val.compare("FALSE")==0 || val.compare("False")==0 || val.compare("false")==0) {
					ALLOWGU = false;
				} else {
					cerr<<"Can not recognize ALLOWGU = "<<val<<endl<<"Please specify TRUE or FALSE in your configuration file (by default: config)."<<endl;
					exit(1);
				}
			}else if(name.compare("MINSTACKLEN")==0){
				iss>>MINSTACKLEN;
			}else if(name.compare("MINLOOPLEN")==0){
				iss>>MINLOOPLEN;
			}else if(name.compare("MINHYDROGEN")==0){
				iss>>MINHYDROGEN;
			}else if(name.compare("MAXSEQLEN")==0){
				iss>>MAXSEQLEN;
//			}else if(name.compare("MAXINWARDOVERLAP_PERCENT")==0) {
//				iss>>MAXINWARDOVERLAP_PERCENT;
//			}else if(name.compare("MAXPARALLELOVERLAP")==0) {
//				iss>>MAXPARALLELOVERLAP;
			}else if(name.compare("MAXOVERLAP")==0) {
				iss>>MAXOVERLAP;
			}else if(name.compare("BSTAR")==0) {
				iss>>BSTAR;
			}else if(name.compare("VAR1")==0) {
				iss>>VAR1;
			}else if(name.compare("VAR2")==0) {
				iss>>VAR2;
			}else if(name.compare("P_MAXGAP")==0) {
				iss>>P_MAXGAP;
			}else if(name.compare("BARRIERCUTOFF")==0) {
				iss>>BARRIERCUTOFF;
			}else if(name.compare("BARRIERCUTOFFDELTA")==0) {
				iss>>BARRIERCUTOFFDELTA;
			}else if(name.compare("MAXMATRIXSZ")==0) {
				iss>>MAXMATRIXSZ;
			}else {
				cerr<<"Can not analyze "<<sline<<"in your configuration file:"<<configfile<<endl<<"Please update it."<<endl;
				exit(1);
			}
		}
	}
}

string Config::trim(const string & in_s) {
	string ret = in_s;
	while(ret[0] == ' ' || ret[0] == '\t'){
		ret = ret.substr(1);
	}
	while(ret[sz(ret)-1] == ' ' || ret[sz(ret)-1] == '\t' 
	   || ret[sz(ret)-1] == '\n') {
		ret = ret.substr(0, sz(ret)-1);
	}
	return ret;
}

void Config::show(ostream & fout) {
	fout<<"ALLOWGU = ";
	if(ALLOWGU)  fout<<"TRUE"<<endl;
	else fout<<"FALSE"<<endl;
	fout<<"MINSTACKLEN = "<<MINSTACKLEN<<endl;
	fout<<"MINLOOPLEN = "<<MINLOOPLEN<<endl;
	fout<<"MINHYDROGEN = "<<MINHYDROGEN<<endl;
	fout<<"MAXSEQLEN = "<<MAXSEQLEN<<endl;
//	fout<<"MAXINWARDOVERLAP_PERCENT = "<<MAXINWARDOVERLAP_PERCENT<<endl;
//	fout<<"MAXPARALLELOVERLAP = "<<MAXPARALLELOVERLAP<<endl;
	fout<<"MAXOVERLAP = "<<MAXOVERLAP<<endl;
	fout<<"MAXMATRIXSZ = "<<MAXMATRIXSZ<<endl;
	fout<<"BARRIERCUTOFF = "<<BARRIERCUTOFF<<endl;
	fout<<"BARRIERCUTOFFDELTA = "<<BARRIERCUTOFFDELTA<<endl;
	fout<<"BSTAR = "<<BSTAR<<endl;
	fout<<"VAR1 = "<<VAR1<<endl;
	fout<<"VAR2 = "<<VAR2<<endl;
}


void Config::increaseBARRIERCUTOFF() {
	BARRIERCUTOFF += BARRIERCUTOFFDELTA;
}

void Config::resetBARRIERCUTOFF(const int & num ) {
	if(num <= 5000) {
	} else if(num <= 10000){ 
		BARRIERCUTOFF += BARRIERCUTOFFDELTA * 1;
	} else if(num <= 20000) {
		BARRIERCUTOFF += BARRIERCUTOFFDELTA * 2;
	} else if(num <= 30000) {
		BARRIERCUTOFF += BARRIERCUTOFFDELTA * 3;
	} else if(num <= 50000) {
		BARRIERCUTOFF += BARRIERCUTOFFDELTA * 4;
	} else if(num <= 70000) {
		BARRIERCUTOFF += BARRIERCUTOFFDELTA * 5;
	} else {
		BARRIERCUTOFF += BARRIERCUTOFFDELTA * 6;
	}
}
